import numpy as np
import mcdc as MCDC
from mcdc import type_
from mcdc.main import closeout
from mcdc.kernel import rng, AxV, FxV, HxV, preconditioner
import mcdc.global_ as mcdc_

input_deck = mcdc_.input_deck


def iqmc_dummy_mcdc_variable():
    """
    This function returns the global "mcdc" container. Inputs are
    taken from the kornreich eigenvalue problem.

    """
    # Set materials
    m1 = MCDC.material(
        capture=np.array([0.0]),
        scatter=np.array([[0.9]]),
        fission=np.array([0.1]),
        nu_p=np.array([6.0]),
    )
    m2 = MCDC.material(
        capture=np.array([0.68]),
        scatter=np.array([[0.2]]),
        fission=np.array([0.12]),
        nu_p=np.array([2.5]),
    )

    # Set surfaces
    s1 = MCDC.surface("plane-x", x=0.0, bc="vacuum")
    s2 = MCDC.surface("plane-x", x=1.5)
    s3 = MCDC.surface("plane-x", x=2.5, bc="vacuum")

    # Set cells
    MCDC.cell([+s1, -s2], m1)
    MCDC.cell([+s2, -s3], m2)

    # =============================================================================
    # iQMC Parameters
    # =============================================================================
    N = 100
    maxit = 10
    tol = 1e-3
    x = np.arange(0.0, 2.6, 0.1)
    Nx = len(x) - 1
    generator = "halton"
    solver = "davidson"
    fixed_source = np.zeros(Nx)
    phi0 = np.ones((Nx))

    # =============================================================================
    # Set tally, setting, and run mcdc
    # =============================================================================

    MCDC.iQMC(
        x=x,
        fixed_source=fixed_source,
        phi0=phi0,
        maxitt=maxit,
        tol=tol,
        generator=generator,
        eigenmode_solver=solver,
    )
    # Setting
    MCDC.setting(N_particle=N)
    MCDC.eigenmode()
    return MCDC.prepare()


def test_rn_basic():
    """
    Basic test routine for random number generator.
    Get the first 5 random numbers, skip a few, get 5 more and compare to
    reference data from [1]

    Seed numbers for index 1-5, 123456-123460

    [1] F. B. Brown, “Random number generation with arbitrary strides”,
        Trans. Am. Nucl. Soc, 71, 202 (1994)

    """
    MCDC.reset_cards()

    ref_data = np.array(
        (
            1,
            2806196910506780710,
            6924308458965941631,
            7093833571386932060,
            4133560638274335821,
        )
    )

    mcdc = iqmc_dummy_mcdc_variable()

    # run through the first five seeds (1-5)
    for i in range(5):
        assert mcdc["setting"]["rng_seed"] == ref_data[i]
        rng(mcdc["setting"])


def test_AxV_linearity():
    """
    AxV is the linear operator used for GMRES in iQMC.

    Linear operators must satisfy conditions of additivity and multiplicity
    defined as:
            - Additivity: f(x+y) = f(x) + f(y)
            - Multiplicity: f(cx) = cf(x)

    We can test both properties with:
            - f(a*x + b*y) = a*f(x) + b*f(y)
    """
    MCDC.reset_cards()

    mcdc = iqmc_dummy_mcdc_variable()

    size = mcdc["technique"]["iqmc"]["total_source"].size
    np.random.seed(123456)
    a = np.random.random()
    b = np.random.random()
    x = np.random.random((size,))
    y = np.random.random((size,))
    rhs = np.zeros((size,))

    F1 = AxV((a * x + b * y), rhs, mcdc)
    F2 = a * AxV(x, rhs, mcdc) + b * AxV(y, rhs, mcdc)
    assert np.allclose(F1, F2, rtol=1e-10)


def test_FxV_linearity():
    """
    FxV is a linear operator used for Davidson in iQMC.

    Linear operators must satisfy conditions of additivity and multiplicity
    defined as:
            - Additivity: f(x+y) = f(x) + f(y)
            - Multiplicity: f(cx) = cf(x)

    We can test both properties with:
            - f(a*x + b*y) = a*f(x) + b*f(y)
    """
    MCDC.reset_cards()

    mcdc = iqmc_dummy_mcdc_variable()

    size = mcdc["technique"]["iqmc"]["total_source"].size
    np.random.seed(123456)
    a = np.random.random()
    b = np.random.random()
    x = np.random.random((size, 1))
    y = np.random.random((size, 1))

    F1 = FxV((a * x + b * y), mcdc)
    F2 = a * FxV(x, mcdc) + b * FxV(y, mcdc)
    assert np.allclose(F1, F2, rtol=1e-10)


def test_HxV_linearity():
    """
    HxV is a linear operator used for Davidson in iQMC.

    Linear operators must satisfy conditions of additivity and multiplicity
    defined as:
            - Additivity: f(x+y) = f(x) + f(y)
            - Multiplicity: f(cx) = cf(x)

    We can test both properties with:
            - f(a*x + b*y) = a*f(x) + b*f(y)
    """
    MCDC.reset_cards()

    mcdc = iqmc_dummy_mcdc_variable()

    size = mcdc["technique"]["iqmc"]["total_source"].size
    np.random.seed(123456)
    a = np.random.random()
    b = np.random.random()
    x = np.random.random((size, 1))
    y = np.random.random((size, 1))

    F1 = HxV((a * x + b * y), mcdc)
    F2 = a * HxV(x, mcdc) + b * HxV(y, mcdc)
    assert np.allclose(F1, F2, rtol=1e-10)


def test_preconditioner_linearity():
    """
    perconditioner is a linear operator used for Davidson in iQMC.

    Linear operators must satisfy conditions of additivity and multiplicity
    defined as:
            - Additivity: f(x+y) = f(x) + f(y)
            - Multiplicity: f(cx) = cf(x)

    We can test both properties with:
            - f(a*x + b*y) = a*f(x) + b*f(y)
    """
    MCDC.reset_cards()

    mcdc = iqmc_dummy_mcdc_variable()

    size = mcdc["technique"]["iqmc"]["total_source"].size
    np.random.seed(123456)
    a = np.random.random()
    b = np.random.random()
    x = np.random.random((size,))
    y = np.random.random((size,))

    F1 = preconditioner((a * x + b * y), mcdc)
    F2 = a * preconditioner(x, mcdc) + b * preconditioner(y, mcdc)
    assert np.allclose(F1, F2, rtol=1e-10)


# if __name__ == "__main__":
# test_rn_basic()
# test_AxV_linearity()
#     test_FxV_linearity()
#     test_HxV_linearity()
#     test_preconditioner_linearity()



import matplotlib.pyplot as plt
import h5py
import sys

from reference import reference


# Load results
output = sys.argv[1]
with h5py.File(output, "r") as f:
    x = f["tally/grid/x"][:]
    dx = x[1:] - x[:-1]
    x_mid = 0.5 * (x[:-1] + x[1:])
    phi = f["tally/flux/mean"][:] / dx * 101.0
    phi_sd = f["tally/flux/sdev"][:] / dx * 101.0

# Reference solution
phi_ref = reference(x)

# Flux - spatial average
plt.plot(x_mid, phi, "-b", label="MC")
plt.fill_between(x_mid, phi - phi_sd, phi + phi_sd, alpha=0.2, color="b")
plt.plot(x_mid, phi_ref, "--r", label="Ref.")
plt.xlabel(r"$x$, cm")
plt.ylabel("Flux")
plt.grid()
plt.legend()
plt.title(r"$\bar{\phi}_i$")
plt.show()



import numpy as np
import matplotlib.pyplot as plt
import h5py
import sys

from reference import reference


# Load results
output = sys.argv[1]
with np.load("SHEM-361.npz") as data:
    E = data["E"]
    G = data["G"]
    E_mid = 0.5 * (E[1:] + E[:-1])
    dE = E[1:] - E[:-1]
with h5py.File(output, "r") as f:
    phi = f["tally/flux/mean"][:] / dE * E_mid
    phi_sd = f["tally/flux/sdev"][:] / dE * E_mid

# Reference solution
phi_ref = reference() / dE * E_mid

# Flux
plt.plot(E_mid, phi, "-b", label="MC")
plt.fill_between(E_mid, phi - phi_sd, phi + phi_sd, alpha=0.2, color="b")
plt.plot(E_mid, phi_ref, "--r", label="analytical")
plt.xscale("log")
plt.xlabel(r"$E$, eV")
plt.ylabel(r"$E\phi(E)$")
plt.grid()
plt.legend()
plt.show()



import numpy as np
import os
import sys

# Fixed source
N_min = 3
N_max = 6
for task in os.scandir("./fixed_source"):
    print(task)
    os.chdir(task)
    os.system("python process.py %i %i" % (N_min, N_max))
    os.chdir(r"../..")

# Eigenvalue
"""
N_min = 1
N_max = 3
for task in os.scandir('./eigenvalue'):
    print(task)
    os.chdir(task)
    os.system("python process.py %i %i"%(N_min,N_max))
    os.chdir(r"../..")
"""



import numpy as np
from numpy import ascontiguousarray as cga
from numba import njit, objmode, jit
from scipy.linalg import eig

import mcdc.kernel as kernel
import mcdc.type_ as type_

import mcdc.print_ as print_module

from mcdc.constant import *
from mcdc.print_ import (
    print_header_batch,
    print_progress,
    print_progress_eigenvalue,
    print_progress_iqmc,
    print_iqmc_eigenvalue_progress,
    print_iqmc_eigenvalue_exit_code,
)


# =========================================================================
# Fixed-source loop
# =========================================================================


@njit
def loop_fixed_source(mcdc):
    # Loop over batches
    for idx_batch in range(mcdc["setting"]["N_batch"]):
        mcdc["idx_batch"] = idx_batch
        seed_batch = kernel.split_seed(idx_batch, mcdc["setting"]["rng_seed"])

        # Print multi-batch header
        if mcdc["setting"]["N_batch"] > 1:
            with objmode():
                print_header_batch(mcdc)
            if mcdc["technique"]["uq"]:
                seed_uq = kernel.split_seed(seed_batch, SEED_SPLIT_UQ)
                kernel.uq_reset(mcdc, seed_uq)

        # Loop over time censuses
        for idx_census in range(mcdc["setting"]["N_census"]):
            mcdc["idx_census"] = idx_census
            seed_census = kernel.split_seed(seed_batch, SEED_SPLIT_CENSUS)

            # Loop over source particles
            seed_source = kernel.split_seed(seed_census, SEED_SPLIT_SOURCE)
            loop_source(seed_source, mcdc)

            # Loop over source precursors
            if mcdc["bank_precursor"]["size"] > 0:
                seed_source_precursor = kernel.split_seed(
                    seed_census, SEED_SPLIT_SOURCE_PRECURSOR
                )
                loop_source_precursor(seed_source_precursor, mcdc)

            # Time census closeout
            if idx_census < mcdc["setting"]["N_census"] - 1:
                # TODO: Output tally (optional)

                # Manage particle banks: population control and work rebalance
                seed_bank = kernel.split_seed(seed_census, SEED_SPLIT_BANK)
                kernel.manage_particle_banks(seed_bank, mcdc)

        # Multi-batch closeout
        if mcdc["setting"]["N_batch"] > 1:
            # Tally history closeout
            kernel.tally_reduce_bin(mcdc)
            kernel.tally_closeout_history(mcdc)
            # Uq closeout
            if mcdc["technique"]["uq"]:
                kernel.uq_tally_closeout_batch(mcdc)

    # Tally closeout
    if mcdc["technique"]["uq"]:
        kernel.uq_tally_closeout(mcdc)
    else:
        kernel.tally_closeout(mcdc)


# =========================================================================
# Eigenvalue loop
# =========================================================================


@njit
def loop_eigenvalue(mcdc):
    # Loop over power iteration cycles
    for idx_cycle in range(mcdc["setting"]["N_cycle"]):
        seed_cycle = kernel.split_seed(idx_cycle, mcdc["setting"]["rng_seed"])

        # Loop over source particles
        seed_source = kernel.split_seed(seed_cycle, SEED_SPLIT_SOURCE)
        loop_source(seed_source, mcdc)

        # Tally "history" closeout
        kernel.eigenvalue_tally_closeout_history(mcdc)
        if mcdc["cycle_active"]:
            kernel.tally_reduce_bin(mcdc)
            kernel.tally_closeout_history(mcdc)

        # Print progress
        with objmode():
            print_progress_eigenvalue(mcdc)

        # Manage particle banks
        seed_bank = kernel.split_seed(seed_cycle, SEED_SPLIT_BANK)
        kernel.manage_particle_banks(seed_bank, mcdc)

        # Entering active cycle?
        mcdc["idx_cycle"] += 1
        if mcdc["idx_cycle"] >= mcdc["setting"]["N_inactive"]:
            mcdc["cycle_active"] = True

    # Tally closeout
    kernel.tally_closeout(mcdc)
    kernel.eigenvalue_tally_closeout(mcdc)


# =============================================================================
# Source loop
# =============================================================================


@njit
def loop_source(seed, mcdc):
    # Progress bar indicator
    N_prog = 0

    # Loop over particle sources
    work_start = mcdc["mpi_work_start"]
    work_size = mcdc["mpi_work_size"]
    work_end = work_start + work_size
    for idx_work in range(work_size):
        seed_work = kernel.split_seed(work_start + idx_work, seed)

        # Particle tracker
        if mcdc["setting"]["track_particle"]:
            mcdc["particle_track_history_ID"] += 1

        # =====================================================================
        # Get a source particle and put into active bank
        # =====================================================================

        # Get from fixed-source?
        if mcdc["bank_source"]["size"] == 0:
            # Sample source
            P = kernel.source_particle(seed_work, mcdc)

        # Get from source bank
        else:
            P = mcdc["bank_source"]["particles"][idx_work]

        # Check if it is beyond current census index
        idx_census = mcdc["idx_census"]
        if P["t"] > mcdc["setting"]["census_time"][idx_census]:
            P["t"] += SHIFT
            kernel.add_particle(P, mcdc["bank_census"])
        else:
            # Add the source particle into the active bank
            kernel.add_particle(P, mcdc["bank_active"])

        # =====================================================================
        # Run the source particle and its secondaries
        # =====================================================================

        # Loop until active bank is exhausted
        while mcdc["bank_active"]["size"] > 0:
            # Get particle from active bank
            P = kernel.get_particle(mcdc["bank_active"], mcdc)

            # Apply weight window
            if mcdc["technique"]["weight_window"]:
                kernel.weight_window(P, mcdc)

            # Particle tracker
            if mcdc["setting"]["track_particle"]:
                mcdc["particle_track_particle_ID"] += 1

            # Particle loop
            loop_particle(P, mcdc)

        # =====================================================================
        # Closeout
        # =====================================================================

        # Tally history closeout for one-batch fixed-source simulation
        if not mcdc["setting"]["mode_eigenvalue"] and mcdc["setting"]["N_batch"] == 1:
            kernel.tally_closeout_history(mcdc)

        # Tally history closeout for multi-batch uq simulation
        if mcdc["technique"]["uq"]:
            kernel.uq_tally_closeout_history(mcdc)

        # Progress printout
        percent = (idx_work + 1.0) / mcdc["mpi_work_size"]
        if mcdc["setting"]["progress_bar"] and int(percent * 100.0) > N_prog:
            N_prog += 1
            with objmode():
                print_progress(percent, mcdc)


# =========================================================================
# Particle loop
# =========================================================================


@njit
def loop_particle(P, mcdc):
    # Particle tracker
    if mcdc["setting"]["track_particle"]:
        kernel.track_particle(P, mcdc)

    while P["alive"]:
        # Find cell from root universe if unknown
        if P["cell_ID"] == -1:
            trans = np.zeros(3)
            P["cell_ID"] = kernel.get_particle_cell(P, 0, trans, mcdc)

        # Determine and move to event
        kernel.move_to_event(P, mcdc)
        event = P["event"]

        # The & operator here is a bitwise and.
        # It is used to determine if an event type is part of the particle event.

        # Collision events
        if event & EVENT_COLLISION:
            # Generate IC?
            if mcdc["technique"]["IC_generator"] and mcdc["cycle_active"]:
                kernel.bank_IC(P, mcdc)

            # Branchless collision?
            if mcdc["technique"]["branchless_collision"]:
                kernel.branchless_collision(P, mcdc)

            # Analog collision
            else:
                # Get collision type
                kernel.collision(P, mcdc)
                event = P["event"]

                # Perform collision
                if event == EVENT_SCATTERING:
                    kernel.scattering(P, mcdc)
                elif event == EVENT_FISSION:
                    kernel.fission(P, mcdc)

                # Sensitivity quantification for nuclide?
                material = mcdc["materials"][P["material_ID"]]
                if material["sensitivity"] and (
                    P["sensitivity_ID"] == 0
                    or mcdc["technique"]["dsm_order"] == 2
                    and P["sensitivity_ID"] <= mcdc["setting"]["N_sensitivity"]
                ):
                    kernel.sensitivity_material(P, mcdc)

        # Surface crossing
        if event & EVENT_SURFACE:
            kernel.surface_crossing(P, mcdc)

        # Lattice or mesh crossing (skipped if surface crossing)
        elif event & EVENT_LATTICE or event & EVENT_MESH:
            kernel.shift_particle(P, SHIFT)

        # Moving surface transition
        if event & EVENT_SURFACE_MOVE:
            P["t"] += SHIFT
            P["cell_ID"] = -1

        # Census time crossing
        if event & EVENT_CENSUS:
            P["t"] += SHIFT
            kernel.add_particle(kernel.copy_particle(P), mcdc["bank_census"])
            P["alive"] = False

        # Time boundary crossing
        if event & EVENT_TIME_BOUNDARY:
            P["alive"] = False

        # Apply weight window
        if P["alive"] and mcdc["technique"]["weight_window"]:
            kernel.weight_window(P, mcdc)

        # Apply weight roulette
        if P["alive"] and mcdc["technique"]["weight_roulette"]:
            # check if weight has fallen below threshold
            if abs(P["w"]) <= mcdc["technique"]["wr_threshold"]:
                kernel.weight_roulette(P, mcdc)

    # Particle tracker
    if mcdc["setting"]["track_particle"]:
        kernel.track_particle(P, mcdc)


# =============================================================================
# iQMC Loops
# =============================================================================


@njit
def loop_iqmc(mcdc):
    # function calls from specified solvers
    iqmc = mcdc["technique"]["iqmc"]
    kernel.iqmc_preprocess(mcdc)
    if mcdc["setting"]["mode_eigenvalue"]:
        if iqmc["eigenmode_solver"] == "davidson":
            davidson(mcdc)
        if iqmc["eigenmode_solver"] == "power_iteration":
            power_iteration(mcdc)
    else:
        if iqmc["fixed_source_solver"] == "source_iteration":
            source_iteration(mcdc)
        if iqmc["fixed_source_solver"] == "gmres":
            gmres(mcdc)


@njit
def source_iteration(mcdc):
    simulation_end = False
    iqmc = mcdc["technique"]["iqmc"]
    total_source_old = iqmc["total_source"].copy()

    while not simulation_end:
        # reset particle bank size
        mcdc["bank_source"]["size"] = 0
        # initialize particles with LDS
        kernel.iqmc_prepare_particles(mcdc)
        # reset tallies for next loop
        kernel.iqmc_reset_tallies(iqmc)
        # sweep particles
        iqmc["sweep_counter"] += 1
        loop_source(0, mcdc)

        # sum resultant flux on all processors
        kernel.iqmc_distribute_tallies(iqmc)
        iqmc["itt"] += 1
        kernel.iqmc_update_source(mcdc)
        # combine source tallies into one vector
        kernel.iqmc_consolidate_sources(mcdc)
        # calculate norm of sources
        iqmc["res"] = kernel.iqmc_res(iqmc["total_source"], total_source_old)
        # iQMC convergence criteria
        if (iqmc["itt"] == iqmc["maxitt"]) or (iqmc["res"] <= iqmc["tol"]):
            simulation_end = True

        # Print progress
        if not mcdc["setting"]["mode_eigenvalue"]:
            with objmode():
                print_progress_iqmc(mcdc)

        # set  source_old = current source
        total_source_old = iqmc["total_source"].copy()


@njit
def gmres(mcdc):
    """
    GMRES solver.
    ----------
    Linear Krylov solver. Solves problem of the form Ax = b.

    References
    ----------
    .. [1] Yousef Saad, "Iterative Methods for Sparse Linear Systems,
        Second Edition", SIAM, pp. 151-172, pp. 272-275, 2003
        http://www-users.cs.umn.edu/~saad/books.html
    .. [2] C. T. Kelley, http://www4.ncsu.edu/~ctk/matlab_roots.html

    code adapted from: https://github.com/pygbe/pygbe/blob/master/pygbe/gmres.py

    """
    iqmc = mcdc["technique"]["iqmc"]
    max_iter = iqmc["maxitt"]
    R = iqmc["krylov_restart"]
    tol = iqmc["tol"]

    fixed_source = iqmc["fixed_source"]
    single_vector = iqmc["fixed_source"].size
    b = np.zeros_like(iqmc["total_source"])
    b[:single_vector] = np.reshape(fixed_source, fixed_source.size)
    X = iqmc["total_source"].copy()
    # initial residual
    r = b - kernel.AxV(X, b, mcdc)
    normr = np.linalg.norm(r)

    # Defining dimension
    dimen = X.size
    # Set number of outer and inner iterations
    if R > dimen:
        # set number of outter iterations to max allowable (A.shape[0])
        R = dimen
    max_inner = R
    xtype = np.float64

    # max_outer should be max_iter/max_inner but this might not be an integer
    # so we get the ceil of the division.
    # In the inner loop there is a if statement to break in case max_iter is
    # reached.
    max_outer = int(np.ceil(max_iter / max_inner))

    # Check initial guess ( scaling by b, if b != 0, must account for
    # case when norm(b) is very small)
    normb = np.linalg.norm(b)
    if normb == 0.0:
        normb = 1.0
    if normr < tol * normb:
        return X, 0

    iteration = 0

    # GMRES starts here
    for outer in range(max_outer):
        # Preallocate for Givens Rotations, Hessenberg matrix and Krylov Space
        Q = []
        H = np.zeros((max_inner + 1, max_inner + 1), dtype=xtype)
        V = np.zeros((max_inner + 1, dimen), dtype=xtype)

        # vs store the pointers to each column of V.
        # This saves a considerable amount of time.
        vs = []
        V[0, :] = (1.0 / normr) * r
        vs.append(V[0, :])

        # Saving initial residual to be used to calculate the rel_resid
        if iteration == 0:
            res_0 = normb

        # RHS vector in the Krylov space
        g = np.zeros((dimen,), dtype=xtype)
        g[0] = normr

        for inner in range(max_inner):
            # New search direction
            v = V[inner + 1, :]
            v[:] = kernel.AxV(vs[-1], b, mcdc)
            vs.append(v)

            # Modified Gram Schmidt
            for k in range(inner + 1):
                vk = vs[k]
                alpha = np.dot(vk, v)
                H[inner, k] = alpha
                v[:] = vk * (-alpha) + v[:]

            normv = np.linalg.norm(v)
            H[inner, inner + 1] = normv

            # Check for breakdown
            if H[inner, inner + 1] != 0.0:
                v[:] = (1.0 / H[inner, inner + 1]) * v

            # Apply for Givens rotations to H
            if inner > 0:
                for j in range(inner):
                    Qloc = Q[j]
                    H[inner, :][j : j + 2] = np.dot(Qloc, H[inner, :][j : j + 2])

            # Calculate and apply next complex-valued Givens rotations

            # If max_inner = dimen, we don't need to calculate, this
            # is unnecessary for the last inner iteration when inner = dimen -1
            if inner != dimen - 1:
                if H[inner, inner + 1] != 0:
                    # Caclulate matrix rotations
                    c, s, _ = kernel.lartg(H[inner, inner], H[inner, inner + 1])
                    Qblock = np.array([[c, s], [-np.conjugate(s), c]], dtype=xtype)
                    Q.append(Qblock)

                    # Apply Givens Rotations to RHS for the linear system in
                    # the krylov space.
                    g[inner : inner + 2] = np.dot(Qblock, g[inner : inner + 2])

                    # Apply Givens rotations to H
                    H[inner, inner] = np.dot(Qblock[0, :], H[inner, inner : inner + 2])
                    H[inner, inner + 1] = 0.0

            iteration += 1

            if inner < max_inner - 1:
                normr = abs(g[inner + 1])
                rel_resid = normr / res_0
                iqmc["res"] = rel_resid

            iqmc["itt"] += 1
            if not mcdc["setting"]["mode_eigenvalue"]:
                with objmode():
                    print_progress_iqmc(mcdc)

            if rel_resid < tol:
                break
            if iqmc["itt"] >= max_iter:
                break

        # end inner loop, back to outer loop
        # Find best update to X in Krylov Space V.  Solve inner X inner system.
        y = np.linalg.solve(H[0 : inner + 1, 0 : inner + 1].T, g[0 : inner + 1])
        update = np.ravel(np.dot(cga(V[: inner + 1, :].T), y.reshape(-1, 1)))
        X = X + update
        aux = kernel.AxV(X, b, mcdc)
        r = b - aux
        normr = np.linalg.norm(r)
        rel_resid = normr / res_0
        iqmc["res"] = rel_resid
        if rel_resid < tol:
            break
        if iqmc["itt"] >= max_iter:
            return


@njit
def power_iteration(mcdc):
    simulation_end = False
    iqmc = mcdc["technique"]["iqmc"]
    # iteration tolerance
    tol = iqmc["tol"]
    # maximum number of iterations
    maxit = iqmc["maxitt"]
    score_bin = iqmc["score"]
    k_old = mcdc["k_eff"]
    solver = iqmc["fixed_source_solver"]

    fission_source_old = score_bin["fission-source"].copy()

    while not simulation_end:
        # iterate over scattering source
        if solver == "source_iteration":
            iqmc["maxitt"] = 1
            source_iteration(mcdc)
            iqmc["maxitt"] = maxit
        if solver == "gmres":
            gmres(mcdc)
        # reset counter for inner iteration
        iqmc["itt"] = 0

        # update k_eff
        mcdc["k_eff"] *= score_bin["fission-source"][0] / fission_source_old[0]

        # calculate diff in keff
        iqmc["res_outter"] = abs(mcdc["k_eff"] - k_old) / k_old
        k_old = mcdc["k_eff"]
        # store outter iteration values
        score_bin["effective-fission-outter"] = score_bin["effective-fission"].copy()
        fission_source_old = score_bin["fission-source"].copy()
        iqmc["itt_outter"] += 1

        with objmode():
            print_iqmc_eigenvalue_progress(mcdc)

        # iQMC convergence criteria
        if (iqmc["itt_outter"] == maxit) or (iqmc["res_outter"] <= tol):
            simulation_end = True
            with objmode():
                print_iqmc_eigenvalue_exit_code(mcdc)


@njit
def davidson(mcdc):
    """
    The generalized Davidson method is a Krylov subspace method for solving
    the generalized eigenvalue problem. The algorithm here is based on the
    outline in:

        Subramanian, C., et al. "The Davidson method as an alternative to
        power iterations for criticality calculations." Annals of nuclear
        energy 38.12 (2011): 2818-2823.

    """
    # TODO: handle imaginary eigenvalues
    iqmc = mcdc["technique"]["iqmc"]
    # Davidson parameters
    simulation_end = False
    maxit = iqmc["maxitt"]
    tol = iqmc["tol"]
    # num_sweeps: number of preconditioner sweeps
    num_sweeps = iqmc["preconditioner_sweeps"]
    # m : restart parameter
    m = iqmc["krylov_restart"]
    k_old = mcdc["k_eff"]
    # initial size of Krylov subspace
    Vsize = 1
    # l : number of eigenvalues to solve for
    l = 1
    # vector size
    Nt = iqmc["total_source"].size
    # allocate memory then use slice indexing in loop
    V = np.zeros((Nt, m), dtype=np.float64)
    HV = np.zeros((Nt, m), dtype=np.float64)
    FV = np.zeros((Nt, m), dtype=np.float64)

    V0 = iqmc["total_source"].copy()
    V0 = kernel.preconditioner(V0, mcdc, num_sweeps=5)
    # orthonormalize initial guess
    V0 = V0 / np.linalg.norm(V0)
    V[:, 0] = V0

    if m is None:
        # unless specified there is no restart parameter
        m = maxit + 1

    # Davidson Routine
    while not simulation_end:
        # Calculate V*H*V (HxV is scattering linear operator function)
        HV[:, Vsize - 1] = kernel.HxV(V[:, :Vsize], mcdc)
        VHV = np.dot(cga(V[:, :Vsize].T), cga(HV[:, :Vsize]))
        # Calculate V*F*V (FxV is fission linear operator function)
        FV[:, Vsize - 1] = kernel.FxV(V[:, :Vsize], mcdc)
        VFV = np.dot(cga(V[:, :Vsize].T), cga(FV[:, :Vsize]))
        # solve for eigenvalues and vectors
        with objmode(Lambda="complex128[:]", w="complex128[:,:]"):
            Lambda, w = eig(VFV, b=VHV)
            Lambda = np.array(Lambda, dtype=np.complex128)
            w = np.array(w, dtype=np.complex128)

        assert Lambda.imag.all() == 0.0
        Lambda = Lambda.real
        w = w.real
        # get indices of eigenvalues from largest to smallest
        idx = np.flip(Lambda.argsort())
        # sort eigenvalues from largest to smallest
        Lambda = Lambda[idx]
        # take the l largest eigenvalues
        Lambda = Lambda[:l]
        # sort corresponding eigenvector (oriented by column)
        w = w[:, idx]
        # take the l largest eigenvectors
        w = w[:, :l]
        # assign keff
        mcdc["k_eff"] = Lambda[0]
        # Ritz vector
        u = np.dot(cga(V[:, :Vsize]), cga(w))
        # residual
        res = kernel.FxV(u, mcdc) - Lambda * kernel.HxV(u, mcdc)
        iqmc["res_outter"] = abs(mcdc["k_eff"] - k_old) / k_old
        k_old = mcdc["k_eff"]
        iqmc["itt_outter"] += 1
        with objmode():
            print_iqmc_eigenvalue_progress(mcdc)

        # check convergence criteria
        if (iqmc["itt_outter"] >= maxit) or (iqmc["res_outter"] <= tol):
            simulation_end = True
            with objmode():
                print_iqmc_eigenvalue_exit_code(mcdc)
            # return
        else:
            # Precondition for next iteration
            t = kernel.preconditioner(res, mcdc, num_sweeps)
            # check restart condition
            if Vsize <= m - l:
                # appends new orthogonalization to V
                V[:, : Vsize + 1] = kernel.modified_gram_schmidt(V[:, :Vsize], t)
                Vsize += 1
            else:
                # "restarts" by appending to a new array
                Vsize = l + 1
                V[:, :Vsize] = kernel.modified_gram_schmidt(u, t)
        # TODO: normalize and save final scalar flux
        iqmc["score"]["flux"] /= iqmc["score"]["flux"].sum()


# =============================================================================
# Precursor source loop
# =============================================================================


@njit
def loop_source_precursor(seed, mcdc):
    # TODO: censussed neutrons seeding is still not reproducible

    # Progress bar indicator
    N_prog = 0

    # =========================================================================
    # Sync. RNG skip ahead for reproducibility
    # =========================================================================

    # Exscan upper estimate of number of particles generated locally
    idx_start, N_local, N_global = kernel.bank_scanning_DNP(
        mcdc["bank_precursor"], mcdc
    )

    # =========================================================================
    # Loop over precursor sources
    # =========================================================================

    for idx_work in range(mcdc["mpi_work_size_precursor"]):
        # Get precursor
        DNP = mcdc["bank_precursor"]["precursors"][idx_work]

        # Set groups
        j = DNP["g"]
        g = DNP["n_g"]

        # Determine number of particles to be generated
        w = DNP["w"]
        N = math.floor(w)
        # "Roulette" the last particle
        seed_work = kernel.split_seed(idx_work, seed)
        if kernel.rng_from_seed(seed_work) < w - N:
            N += 1
        DNP["w"] = N

        # =====================================================================
        # Loop over source particles from the source precursor
        # =====================================================================

        for particle_idx in range(N):
            # Create new particle
            P_new = np.zeros(1, dtype=type_.particle)[0]
            part_seed = kernel.split_seed(particle_idx, seed_work)
            P_new["rng_seed"] = part_seed
            P_new["alive"] = True
            P_new["w"] = 1.0
            P_new["sensitivity_ID"] = 0

            # Set position
            P_new["x"] = DNP["x"]
            P_new["y"] = DNP["y"]
            P_new["z"] = DNP["z"]

            # Get material
            trans = np.zeros(3)
            P_new["cell_ID"] = kernel.get_particle_cell(P_new, 0, trans, mcdc)
            material_ID = kernel.get_particle_material(P_new, mcdc)
            material = mcdc["materials"][material_ID]
            G = material["G"]

            # Sample nuclide and get spectrum and decay constant
            N_nuclide = material["N_nuclide"]
            if N_nuclide == 1:
                nuclide = mcdc["nuclides"][material["nuclide_IDs"][0]]
                spectrum = nuclide["chi_d"][j]
                decay = nuclide["decay"][j]
            else:
                SigmaF = material["fission"][g]  # MG only
                nu_d = material["nu_d"][g]
                xi = kernel.rng(P_new) * nu_d[j] * SigmaF
                tot = 0.0
                for i in range(N_nuclide):
                    nuclide = mcdc["nuclides"][material["nuclide_IDs"][i]]
                    density = material["nuclide_densities"][i]
                    tot += density * nuclide["nu_d"][g, j] * nuclide["fission"][g]
                    if xi < tot:
                        # Nuclide determined, now get the constant and spectruum
                        spectrum = nuclide["chi_d"][j]
                        decay = nuclide["decay"][j]
                        break

            # Sample emission time
            P_new["t"] = -math.log(kernel.rng(P_new)) / decay
            idx_census = mcdc["idx_census"]
            if idx_census > 0:
                P_new["t"] += mcdc["setting"]["census_time"][idx_census - 1]

            # Accept if it is inside current census index
            if P_new["t"] < mcdc["setting"]["census_time"][idx_census]:
                # Reduce precursor weight
                DNP["w"] -= 1.0

                # Skip if it's beyond time boundary
                if P_new["t"] > mcdc["setting"]["time_boundary"]:
                    continue

                # Sample energy
                xi = kernel.rng(P_new)
                tot = 0.0
                for g_out in range(G):
                    tot += spectrum[g_out]
                    if tot > xi:
                        break
                P_new["g"] = g_out

                # Sample direction
                (
                    P_new["ux"],
                    P_new["uy"],
                    P_new["uz"],
                ) = kernel.sample_isotropic_direction(P_new)

                # Push to active bank
                kernel.add_particle(kernel.copy_particle(P_new), mcdc["bank_active"])

                # Loop until active bank is exhausted
                while mcdc["bank_active"]["size"] > 0:
                    # Get particle from active bank
                    P = kernel.get_particle(mcdc["bank_active"], mcdc)

                    # Apply weight window
                    if mcdc["technique"]["weight_window"]:
                        kernel.weight_window(P, mcdc)

                    # Particle tracker
                    if mcdc["setting"]["track_particle"]:
                        mcdc["particle_track_particle_ID"] += 1

                    # Particle loop
                    loop_particle(P, mcdc)

        # =====================================================================
        # Closeout
        # =====================================================================

        # Tally history closeout for fixed-source simulation
        if not mcdc["setting"]["mode_eigenvalue"]:
            kernel.tally_closeout_history(mcdc)

        # Progress printout
        percent = (idx_work + 1.0) / mcdc["mpi_work_size_precursor"]
        if mcdc["setting"]["progress_bar"] and int(percent * 100.0) > N_prog:
            N_prog += 1
            with objmode():
                print_progress(percent, mcdc)



import numpy as np
import h5py
import sys
import matplotlib.pyplot as plt
import matplotlib.animation as animation

sys.path.append("../../")
import tool


# Cases run
N_min = int(sys.argv[1])
N_max = int(sys.argv[2])
N_particle_list = np.logspace(N_min, N_max, (N_max - N_min) * 2 + 1)

# Reference solution
data = np.load("reference.npz")
phi_ref_ = data["phi"].T
n_ref = data["n"].T

# Error container
error = np.zeros(len(N_particle_list))
error_n = np.zeros(len(N_particle_list))

error_max = np.zeros(len(N_particle_list))
error_max_n = np.zeros(len(N_particle_list))

# Calculate error
for i, N_particle in enumerate(N_particle_list):
    # Get results
    with np.load("SHEM-361.npz") as data:
        E = data["E"]
        G = data["G"]
        speed = data["v"]
        E_mid = 0.5 * (E[1:] + E[:-1])
        dE = E[1:] - E[:-1]
    with h5py.File("output_%i.h5" % int(N_particle), "r") as f:
        t = f["tally/grid/t"][:]
        dt = t[1:] - t[:-1]
        K = len(t) - 1
        phi = f["tally/flux/mean"][:]

    # Neutron density
    n = np.zeros(K)
    for k in range(K):
        n[k] = np.sum(phi[:, k] / speed) / dt[k]

    # Normalize
    phi_ref = np.zeros_like(phi_ref_)
    for k in range(K):
        phi_ref[:, k] = phi_ref_[:, k] * E_mid / dE
        phi[:, k] *= E_mid / dE / dt[k]

    error[i] = tool.rerror(phi, phi_ref)
    error_n[i] = tool.rerror(n, n_ref)
    error_max[i] = tool.rerror_max(phi, phi_ref)
    error_max_n[i] = tool.rerror_max(n, n_ref)

tool.plot_convergence("inf_shem361_td_flux", N_particle_list, error, error_max)
tool.plot_convergence("inf_shem361_td_n", N_particle_list, error_n, error_max_n)



import numpy as np
import h5py

import mcdc


# =============================================================================
# Set model and run
# =============================================================================

m1 = mcdc.material(capture=np.array([0.90]))
m2 = mcdc.material(capture=np.array([0.15]))
m3 = mcdc.material(capture=np.array([0.60]))

s1 = mcdc.surface("plane-x", x=-1.0, bc="vacuum")
s2 = mcdc.surface("plane-x", x=2.0)
s3 = mcdc.surface("plane-x", x=5.0)
s4 = mcdc.surface("plane-x", x=6.0, bc="vacuum")

mcdc.cell([+s1, -s2], m1)
mcdc.cell([+s2, -s3], m2)
mcdc.cell([+s3, -s4], m3)

mcdc.source(point=[0.0, 0.0, 0.0], direction=[1.0, 0.0, 0.0])

scores = ["exit"]
mcdc.tally(scores=scores, x=np.linspace(0.0, 6.0, 2))

mcdc.setting(N_particle=1e1, N_batch=1e1, progress_bar=False)

mcdc.uq(material=m1, distribution="uniform", capture=np.array([0.7]))
mcdc.uq(material=m2, distribution="uniform", capture=np.array([0.12]))
mcdc.uq(material=m3, distribution="uniform", capture=np.array([0.5]))

mcdc.run()

# =========================================================================
# Check output
# =========================================================================

output = h5py.File("output.h5", "r")
answer = h5py.File("answer.h5", "r")
for score in scores:
    name = "tally/" + score + "/mean"
    a = output[name][:]
    b = answer[name][:]
    assert np.isclose(a, b).all()

    name = "tally/" + score + "/sdev"
    a = output[name][:]
    b = answer[name][:]
    assert np.isclose(a, b).all()

    name = "tally/" + score + "/uq_var"
    a = output[name][:]
    b = answer[name][:]
    assert np.isclose(a, b).all()

output.close()
answer.close()



# User interface
from mcdc.input_ import (
    nuclide,
    material,
    surface,
    cell,
    universe,
    lattice,
    source,
    tally,
    setting,
    eigenmode,
    implicit_capture,
    weighted_emission,
    population_control,
    branchless_collision,
    time_census,
    weight_window,
    iQMC,
    weight_roulette,
    IC_generator,
    dsm,
    uq,
    print_card,
    reset_cards,
)
from mcdc.main import run, prepare
from mcdc.visualizer import visualize



import math, sys, os, h5py
import numpy as np

from mpi4py import MPI

# Basic types
float64 = np.float64
int64 = np.int64
int32 = np.int32
uint64 = np.uint64
bool_ = np.bool_
str_ = "U30"

# MC/DC types, will be defined based on input deck
particle = None
particle_record = None
nuclide = None
material = None
surface = None
cell = None
universe = None
lattice = None
source = None
setting = None
tally = None
technique = None
global_ = None


# ==============================================================================
# Particle
# ==============================================================================


# Particle (in-flight)
def make_type_particle(input_deck):
    global particle

    struct = [
        ("x", float64),
        ("y", float64),
        ("z", float64),
        ("t", float64),
        ("ux", float64),
        ("uy", float64),
        ("uz", float64),
        ("g", uint64),
        ("E", float64),
        ("w", float64),
        ("alive", bool_),
        ("material_ID", int64),
        ("cell_ID", int64),
        ("surface_ID", int64),
        ("translation", float64, (3,)),
        ("event", int64),
        ("sensitivity_ID", int64),
        ("rng_seed", uint64),
    ]

    # Get modes
    iQMC = input_deck.technique["iQMC"]

    # =========================================================================
    # iQMC
    # =========================================================================

    # Default number of groups for iQMC
    G = 1

    # iQMC vector of weights
    if iQMC:
        G = input_deck.materials[0]["G"]
    iqmc_struct = [("w", float64, (G,))]
    struct += [("iqmc", iqmc_struct)]

    # Save type
    particle = np.dtype(struct)


# Particle record (in-bank)
def make_type_particle_record(input_deck):
    global particle_record

    struct = [
        ("x", float64),
        ("y", float64),
        ("z", float64),
        ("t", float64),
        ("ux", float64),
        ("uy", float64),
        ("uz", float64),
        ("g", uint64),
        ("E", float64),
        ("w", float64),
        ("sensitivity_ID", int64),
        ("rng_seed", uint64),
    ]

    # Get modes
    iQMC = input_deck.technique["iQMC"]

    # =========================================================================
    # iQMC
    # =========================================================================

    # Default number of groups for iQMC
    G = 1

    # iQMC vector of weights
    if iQMC:
        G = input_deck.materials[0]["G"]
    iqmc_struct = [("w", float64, (G,))]
    struct += [("iqmc", iqmc_struct)]

    # Save type
    particle_record = np.dtype(struct)


precursor = np.dtype(
    [
        ("x", float64),
        ("y", float64),
        ("z", float64),
        ("g", uint64),
        ("n_g", uint64),
        ("w", float64),
    ]
)


# ==============================================================================
# Particle bank
# ==============================================================================


def particle_bank(max_size):
    return np.dtype(
        [("particles", particle_record, (max_size,)), ("size", int64), ("tag", "U10")]
    )


def precursor_bank(max_size):
    return np.dtype(
        [("precursors", precursor, (max_size,)), ("size", int64), ("tag", "U10")]
    )


# ==============================================================================
# Nuclide
# ==============================================================================


def make_type_nuclide(input_deck):
    global nuclide

    # Get modes
    mode_CE = input_deck.setting["mode_CE"]
    mode_MG = input_deck.setting["mode_MG"]

    # Get CE sizes
    if mode_CE:
        # Zeros for MG sizes
        G = 1
        J = 0

        # Get maximum energy grid sizes for CE data
        NE_xs = 0
        NE_nu_p = 0
        NE_nu_d = 0
        NE_chi_p = 0
        NE_chi_d1 = 0
        NE_chi_d2 = 0
        NE_chi_d3 = 0
        NE_chi_d4 = 0
        NE_chi_d5 = 0
        NE_chi_d6 = 0

        dir_name = os.getenv("MCDC_XSLIB")
        for nuc in input_deck.nuclides:
            with h5py.File(dir_name + "/" + nuc["name"] + ".h5", "r") as f:
                NE_xs = max(NE_xs, len(f["E_xs"][:]))
                NE_nu_p = max(NE_nu_p, len(f["E_nu_p"][:]))
                NE_nu_d = max(NE_nu_d, len(f["E_nu_d"][:]))
                NE_chi_p = max(NE_chi_p, len(f["E_chi_p"][:]))
                NE_chi_d1 = max(NE_chi_d1, len(f["E_chi_d1"][:]))
                NE_chi_d2 = max(NE_chi_d2, len(f["E_chi_d2"][:]))
                NE_chi_d3 = max(NE_chi_d3, len(f["E_chi_d3"][:]))
                NE_chi_d4 = max(NE_chi_d4, len(f["E_chi_d4"][:]))
                NE_chi_d5 = max(NE_chi_d5, len(f["E_chi_d5"][:]))
                NE_chi_d6 = max(NE_chi_d6, len(f["E_chi_d6"][:]))

    # Get MG sizes
    if mode_MG:
        G = input_deck.materials[0]["G"]
        J = input_deck.materials[0]["J"]

        # Zeros for CE sizes
        NE_xs = 0
        NE_nu_p = 0
        NE_nu_d = 0
        NE_chi_p = 0
        NE_chi_d1 = 0
        NE_chi_d2 = 0
        NE_chi_d3 = 0
        NE_chi_d4 = 0
        NE_chi_d5 = 0
        NE_chi_d6 = 0

    # General data
    struct = [
        ("ID", int64),
        ("fissionable", bool_),
        ("sensitivity", bool_),
        ("sensitivity_ID", int64),
        ("dsm_Np", float64),
        ("uq", bool_),
    ]

    # MG data
    struct += [
        ("G", int64),
        ("J", int64),
        ("speed", float64, (G,)),
        ("decay", float64, (J,)),
        ("total", float64, (G,)),
        ("capture", float64, (G,)),
        ("scatter", float64, (G,)),
        ("fission", float64, (G,)),
        ("nu_s", float64, (G,)),
        ("nu_f", float64, (G,)),
        ("nu_p", float64, (G,)),
        ("nu_d", float64, (G, J)),
        ("chi_s", float64, (G, G)),
        ("chi_p", float64, (G, G)),
        ("chi_d", float64, (J, G)),
    ]

    # CE data
    struct += [
        ("A", float64),
        ("NE_xs", int64),
        ("NE_nu_p", int64),
        ("NE_nu_d", int64),
        ("NE_chi_p", int64),
        ("NE_chi_d1", int64),
        ("NE_chi_d2", int64),
        ("NE_chi_d3", int64),
        ("NE_chi_d4", int64),
        ("NE_chi_d5", int64),
        ("NE_chi_d6", int64),
        ("E_xs", float64, (NE_xs,)),
        ("E_nu_p", float64, (NE_nu_p,)),
        ("E_nu_d", float64, (NE_nu_d,)),
        ("E_chi_p", float64, (NE_chi_p,)),
        ("E_chi_d1", float64, (NE_chi_d1,)),
        ("E_chi_d2", float64, (NE_chi_d2,)),
        ("E_chi_d3", float64, (NE_chi_d3,)),
        ("E_chi_d4", float64, (NE_chi_d4,)),
        ("E_chi_d5", float64, (NE_chi_d5,)),
        ("E_chi_d6", float64, (NE_chi_d6,)),
        ("ce_total", float64, (NE_xs,)),
        ("ce_capture", float64, (NE_xs,)),
        ("ce_scatter", float64, (NE_xs,)),
        ("ce_fission", float64, (NE_xs,)),
        ("ce_nu_p", float64, (NE_nu_p,)),
        ("ce_nu_d", float64, (6, NE_nu_d)),
        ("ce_chi_p", float64, (NE_chi_p,)),
        ("ce_chi_d1", float64, (NE_chi_d1,)),
        ("ce_chi_d2", float64, (NE_chi_d2,)),
        ("ce_chi_d3", float64, (NE_chi_d3,)),
        ("ce_chi_d4", float64, (NE_chi_d4,)),
        ("ce_chi_d5", float64, (NE_chi_d5,)),
        ("ce_chi_d6", float64, (NE_chi_d6,)),
        ("ce_decay", float64, (6,)),
    ]

    # Set the type
    nuclide = np.dtype(struct)


# ==============================================================================
# Material
# ==============================================================================


def make_type_material(input_deck):
    global material

    # Maximum number of nuclides per material
    Nmax_nuclide = max([material["N_nuclide"] for material in input_deck.materials])

    # Get modes
    mode_CE = input_deck.setting["mode_CE"]
    mode_MG = input_deck.setting["mode_MG"]

    # Get CE sizes
    if mode_CE:
        # Zeros for MG sizes
        G = 1
        J = 0

    # Get MG sizes
    if mode_MG:
        G = input_deck.materials[0]["G"]
        J = input_deck.materials[0]["J"]

    # General data
    struct = [
        ("ID", int64),
        ("N_nuclide", int64),
        ("sensitivity", bool_),
        ("nuclide_IDs", int64, (Nmax_nuclide,)),
        ("nuclide_densities", float64, (Nmax_nuclide,)),
        ("uq", bool_),
    ]

    # MG data
    struct += [
        ("G", int64),
        ("J", int64),
        ("speed", float64, (G,)),
        ("total", float64, (G,)),
        ("capture", float64, (G,)),
        ("scatter", float64, (G,)),
        ("fission", float64, (G,)),
        ("nu_s", float64, (G,)),
        ("nu_f", float64, (G,)),
        ("nu_p", float64, (G,)),
        ("nu_d", float64, (G, J)),
        ("chi_s", float64, (G, G)),
        ("chi_p", float64, (G, G)),
    ]

    # Set the type
    material = np.dtype(struct)


# ==============================================================================
# Surface
# ==============================================================================


def make_type_surface(input_deck):
    global surface

    # Maximum number of time-dependent surface slices
    Nmax_slice = 0
    for surface in input_deck.surfaces:
        Nmax_slice = max(Nmax_slice, surface["N_slice"])

    surface = np.dtype(
        [
            ("ID", int64),
            ("N_slice", int64),
            ("vacuum", bool_),
            ("reflective", bool_),
            ("A", float64),
            ("B", float64),
            ("C", float64),
            ("D", float64),
            ("E", float64),
            ("F", float64),
            ("G", float64),
            ("H", float64),
            ("I", float64),
            ("J", float64, (Nmax_slice, 2)),
            ("t", float64, (Nmax_slice + 1,)),
            ("linear", bool_),
            ("nx", float64),
            ("ny", float64),
            ("nz", float64),
            ("sensitivity", bool_),
            ("sensitivity_ID", int64),
            ("dsm_Np", float64),
        ]
    )


# ==============================================================================
# Cell
# ==============================================================================


def make_type_cell(input_deck):
    global cell

    # Maximum number of surfaces per cell
    Nmax_surface = max([cell["N_surface"] for cell in input_deck.cells])

    cell = np.dtype(
        [
            ("ID", int64),
            ("N_surface", int64),
            ("surface_IDs", int64, (Nmax_surface,)),
            ("positive_flags", bool_, (Nmax_surface,)),
            ("material_ID", int64),
            ("lattice", bool_),
            ("lattice_ID", int64),
            ("lattice_center", float64, (3,)),
        ]
    )


# ==============================================================================
# Universe
# ==============================================================================


def make_type_universe(input_deck):
    global universe

    # Maximum number of cells per universe
    Nmax_cell = max([universe["N_cell"] for universe in input_deck.universes])

    # Default root universe, if not defined
    N_cell = len(input_deck.cells)
    N_universe = len(input_deck.universes)
    if N_universe == 1:
        Nmax_cell = N_cell
        card = input_deck.universes[0]
        card["N_cell"] = N_cell
        card["cell_IDs"] = np.arange(N_cell)

    universe = np.dtype(
        [("ID", int64), ("N_cell", int64), ("cell_IDs", int64, (Nmax_cell,))]
    )


# ==============================================================================
# Lattice
# ==============================================================================


mesh_uniform = np.dtype(
    [
        ("x0", float64),
        ("dx", float64),
        ("Nx", int64),
        ("y0", float64),
        ("dy", float64),
        ("Ny", int64),
        ("z0", float64),
        ("dz", float64),
        ("Nz", int64),
    ]
)


def make_type_lattice(input_deck):
    global lattice

    # Max dimensional grids
    Nmax_x = 0
    Nmax_y = 0
    Nmax_z = 0
    for card in input_deck.lattices:
        Nmax_x = max(Nmax_x, card["mesh"]["Nx"])
        Nmax_y = max(Nmax_y, card["mesh"]["Ny"])
        Nmax_z = max(Nmax_z, card["mesh"]["Nz"])

    lattice = np.dtype(
        [("mesh", mesh_uniform), ("universe_IDs", int64, (Nmax_x, Nmax_y, Nmax_z))]
    )


# ==============================================================================
# Source
# ==============================================================================


def make_type_source(input_deck):
    global source

    # Get modes
    mode_CE = input_deck.setting["mode_CE"]
    mode_MG = input_deck.setting["mode_MG"]

    # Get energy data size
    if mode_CE:
        G = 1
        # Maximum number of data point in energy pdf
        Nmax_E = max([source["energy"].shape[1] for source in input_deck.sources])
    if mode_MG:
        G = input_deck.materials[0]["G"]
        Nmax_E = 2

    # General data
    struct = [
        ("ID", int64),
        ("box", bool_),
        ("isotropic", bool_),
        ("white", bool_),
        ("x", float64),
        ("y", float64),
        ("z", float64),
        ("box_x", float64, (2,)),
        ("box_y", float64, (2,)),
        ("box_z", float64, (2,)),
        ("ux", float64),
        ("uy", float64),
        ("uz", float64),
        ("white_x", float64),
        ("white_y", float64),
        ("white_z", float64),
        ("time", float64, (2,)),
        ("prob", float64),
    ]

    # MG data
    struct += [
        ("group", float64, (G,)),
    ]

    # CE data
    struct += [
        ("energy", float64, (2, Nmax_E)),
    ]

    source = np.dtype(struct)


# ==============================================================================
# Tally
# ==============================================================================


# Score lists
score_list = (
    "flux",
    "density",
    "fission",
    "total",
    "current",
    "eddington",
    "exit",
)


def make_type_tally(input_deck):
    global tally

    # Number of sensitivitys parameters
    N_sensitivity = input_deck.setting["N_sensitivity"]

    # Number of tally scores
    Ns = 1 + N_sensitivity
    if input_deck.technique["dsm_order"] == 2:
        Ns = 1 + 2 * N_sensitivity + int(0.5 * N_sensitivity * (N_sensitivity - 1))

    # Get card
    card = input_deck.tally

    # Tally estimator flags
    struct = [("tracklength", bool_)]

    def make_type_score(shape):
        return np.dtype(
            [
                ("bin", float64, shape),
                ("mean", float64, shape),
                ("sdev", float64, shape),
            ]
        )

    # Mesh
    mesh, Nx, Ny, Nz, Nt, Nmu, N_azi, Ng = make_type_mesh(card["mesh"])
    struct += [("mesh", mesh)]

    # Scores and shapes
    scores_shapes = [
        ["flux", (Ns, Ng, Nt, Nx, Ny, Nz, Nmu, N_azi)],
        ["density", (Ns, Ng, Nt, Nx, Ny, Nz, Nmu, N_azi)],
        ["fission", (Ns, Ng, Nt, Nx, Ny, Nz, Nmu, N_azi)],
        ["total", (Ns, Ng, Nt, Nx, Ny, Nz, Nmu, N_azi)],
        ["current", (Ns, Ng, Nt, Nx, Ny, Nz, 3)],
        ["eddington", (Ns, Ng, Nt, Nx, Ny, Nz, 6)],
        ["exit", (Ns, Ng, Nt, 2, Ny, Nz, Nmu, N_azi)],
    ]

    # Add score flags to structure
    for i in range(len(scores_shapes)):
        name = scores_shapes[i][0]
        struct += [(name, bool_)]

    # Add scores to structure
    scores_struct = []
    for i in range(len(scores_shapes)):
        name = scores_shapes[i][0]
        shape = scores_shapes[i][1]
        if not card[name]:
            shape = (0,) * len(shape)
        scores_struct += [(name, make_type_score(shape))]
    scores = np.dtype(scores_struct)
    struct += [("score", scores)]

    # Make tally structure
    tally = np.dtype(struct)


# ==============================================================================
# Setting
# ==============================================================================


def make_type_setting(deck):
    global setting

    card = deck.setting
    struct = [
        # Basic MC simulation parameters
        ("N_particle", uint64),
        ("N_batch", uint64),
        ("rng_seed", uint64),
        ("time_boundary", float64),
        # Physics flags
        ("mode_MG", bool_),
        ("mode_CE", bool_),
        # Misc.
        ("progress_bar", bool_),
        ("output_name", "U30"),
        ("save_input_deck", bool_),
        ("track_particle", bool_),
        # Eigenvalue mode
        ("mode_eigenvalue", bool_),
        ("k_init", float64),
        ("N_inactive", uint64),
        ("N_active", uint64),
        ("N_cycle", uint64),
        ("save_particle", bool_),
        ("gyration_radius", bool_),
        ("gyration_radius_type", uint64),
        # Time census
        ("N_census", uint64),
        ("census_time", float64, (card["N_census"],)),
        # Particle source file
        ("source_file", bool_),
        ("source_file_name", "U30"),
        # Initial condition source file
        ("IC_file", bool_),
        ("IC_file_name", "U30"),
        ("N_precursor", uint64),
        # TODO: Move to technique
        ("N_sensitivity", uint64),
    ]

    # Finalize setting type
    setting = np.dtype(struct)


# ==============================================================================
# Technique
# ==============================================================================

iqmc_score_list = (
    "flux",
    "effective-scattering",
    "effective-fission",
    "tilt-x",
    "tilt-y",
    "tilt-z",
    "tilt-xy",
    "tilt-xz",
    "tilt-yz",
    "fission-power",
    "fission-source",
)


def make_type_technique(input_deck):
    global technique

    # Get sizes
    N_particle = input_deck.setting["N_particle"]

    # Get modes
    mode_MG = input_deck.setting["mode_MG"]

    # Get card
    card = input_deck.technique

    # Number of groups
    if mode_MG:
        G = input_deck.materials[0]["G"]
    else:
        G = 1

    # Technique flags
    struct = [
        ("weighted_emission", bool_),
        ("implicit_capture", bool_),
        ("population_control", bool_),
        ("weight_window", bool_),
        ("weight_roulette", bool_),
        ("iQMC", bool_),
        ("IC_generator", bool_),
        ("branchless_collision", bool_),
        ("uq", bool_),
    ]

    # =========================================================================
    # Population control
    # =========================================================================

    struct += [("pct", int64), ("pc_factor", float64)]

    # =========================================================================
    # Weight window
    # =========================================================================

    # Mesh
    mesh, Nx, Ny, Nz, Nt, Nmu, N_azi, Ng = make_type_mesh(card["ww_mesh"])
    struct += [("ww_mesh", mesh)]
    struct += [("ww_width", float64)]

    # Window
    struct += [("ww", float64, (Nt, Nx, Ny, Nz))]

    # =========================================================================
    # Weight Roulette
    # =========================================================================

    # Constants
    struct += [("wr_threshold", float64), ("wr_chance", float64)]

    # =========================================================================
    # Quasi Monte Carlo
    # =========================================================================
    iqmc_list = []

    # Mesh (for qmc source tallies)
    if card["iQMC"]:
        mesh, Nx, Ny, Nz, Nt, Nmu, N_azi = make_type_mesh_(card["iqmc"]["mesh"])
        Ng = G
        N_dim = 6  # group, x, y, z, mu, phi
    else:
        Nx = Ny = Nz = Nt = Nmu = N_azi = N_particle = Ng = N_dim = 0

    iqmc_list += [("mesh", mesh)]

    # Low-discprenecy sequence
    size = MPI.COMM_WORLD.Get_size()
    rank = MPI.COMM_WORLD.Get_rank()
    # Evenly distribute work
    work_size = math.floor(N_particle / size)
    # Count reminder
    rem = N_particle % size
    # Assign reminder and update starting index
    if rank < rem:
        work_size += 1

    iqmc_list += [("lds", float64, (work_size, N_dim))]
    iqmc_list += [("fixed_source", float64, (Ng, Nt, Nx, Ny, Nz))]
    # TODO: make matidx int32
    iqmc_list += [("material_idx", int64, (Nt, Nx, Ny, Nz))]
    # this is the original source matrix size + all tilted sources
    iqmc_list += [("source", float64, (Ng, Nt, Nx, Ny, Nz))]
    total_size = (Ng * Nt * Nx * Ny * Nz) * card["iqmc"]["krylov_vector_size"]
    iqmc_list += [(("total_source"), float64, (total_size,))]

    # Scores and shapes
    scores_shapes = [
        ["flux", (Ng, Nt, Nx, Ny, Nz)],
        ["effective-scattering", (Ng, Nt, Nx, Ny, Nz)],
        ["effective-fission", (Ng, Nt, Nx, Ny, Nz)],
        ["tilt-x", (Ng, Nt, Nx, Ny, Nz)],
        ["tilt-y", (Ng, Nt, Nx, Ny, Nz)],
        ["tilt-z", (Ng, Nt, Nx, Ny, Nz)],
        ["tilt-xy", (Ng, Nt, Nx, Ny, Nz)],
        ["tilt-xz", (Ng, Nt, Nx, Ny, Nz)],
        ["tilt-yz", (Ng, Nt, Nx, Ny, Nz)],
        ["fission-power", (Ng, Nt, Nx, Ny, Nz)],  # SigmaF*phi
        ["fission-source", (1,)],  # nu*SigmaF*phi
    ]

    if card["iQMC"]:
        if setting["mode_eigenvalue"]:
            if card["iqmc"]["eigenmode_solver"] == "power_iteration":
                card["iqmc"]["score_list"]["fission-source"] = True

    # Add score flags to structure
    score_list = []
    for i in range(len(scores_shapes)):
        name = scores_shapes[i][0]
        score_list += [(name, bool_)]
    score_list = np.dtype(score_list)
    iqmc_list += [("score_list", score_list)]

    # Add scores to structure
    scores_struct = []
    for i in range(len(scores_shapes)):
        name = scores_shapes[i][0]
        shape = scores_shapes[i][1]
        if not card["iqmc"]["score_list"][name]:
            shape = (0,) * len(shape)
        scores_struct += [(name, float64, shape)]
    # TODO: make outter effective fission size zero if not eigenmode
    # (causes problems with numba)
    scores_struct += [("effective-fission-outter", float64, (Ng, Nt, Nx, Ny, Nz))]
    scores = np.dtype(scores_struct)
    iqmc_list += [("score", scores)]

    # Constants
    iqmc_list += [
        ("maxitt", int64),
        ("tol", float64),
        ("itt", int64),
        ("itt_outter", int64),
        ("res", float64),
        ("res_outter", float64),
        ("N_dim", int64),
        ("scramble", bool_),
        ("seed", int64),
        ("generator", str_),
        ("fixed_source_solver", str_),
        ("eigenmode_solver", str_),
        ("krylov_restart", int64),
        ("preconditioner_sweeps", int64),
        ("sweep_counter", int64),
        ("w_min", float64),
    ]

    struct += [("iqmc", iqmc_list)]

    # =========================================================================
    # IC generator
    # =========================================================================

    # Create bank types
    #   We need local banks to ensure reproducibility regardless of # of MPIs
    #   TODO: Having smaller bank buffer (~N_target/MPI_size) and even smaller
    #         local bank would be more efficient.
    if card["IC_generator"]:
        Nn = int(card["IC_N_neutron"] * 1.2)
        Np = int(card["IC_N_precursor"] * 1.2)
        Nn_local = Nn
        Np_local = Np
    else:
        Nn = 0
        Np = 0
        Nn_local = 0
        Np_local = 0
    bank_neutron = particle_bank(Nn)
    bank_neutron_local = particle_bank(Nn_local)
    bank_precursor = precursor_bank(Np)
    bank_precursor_local = precursor_bank(Np_local)

    # The parameters
    struct += [
        ("IC_N_neutron", int64),
        ("IC_N_precursor", int64),
        ("IC_neutron_density", float64),
        ("IC_neutron_density_max", float64),
        ("IC_precursor_density", float64),
        ("IC_precursor_density_max", float64),
        ("IC_bank_neutron_local", bank_neutron_local),
        ("IC_bank_precursor_local", bank_precursor_local),
        ("IC_bank_neutron", bank_neutron),
        ("IC_bank_precursor", bank_precursor),
        ("IC_fission_score", float64),
        ("IC_fission", float64),
    ]

    # =========================================================================
    # Derivative Source Method
    # =========================================================================

    struct += [
        ("dsm_order", int64),
    ]

    # =========================================================================
    # Variance Deconvolution
    # =========================================================================
    struct += [("uq_tally", uq_tally), ("uq_", uq)]

    # Finalize technique type
    technique = np.dtype(struct)


# UQ
def make_type_uq_tally(input_deck):
    global uq_tally

    def make_type_uq_score(shape):
        return np.dtype(
            [
                ("batch_bin", float64, shape),
                ("batch_var", float64, shape),
            ]
        )

    # Tally estimator flags
    struct = []

    # Number of tally scores
    Ns = 1 + input_deck.setting["N_sensitivity"]

    # Tally card
    tally_card = input_deck.tally

    # Mesh, but doesn't need to be added
    mesh, Nx, Ny, Nz, Nt, Nmu, N_azi, Ng = make_type_mesh(tally_card["mesh"])

    # Scores and shapes
    scores_shapes = [
        ["flux", (Ns, Ng, Nt, Nx, Ny, Nz, Nmu, N_azi)],
        ["density", (Ns, Ng, Nt, Nx, Ny, Nz, Nmu, N_azi)],
        ["fission", (Ns, Ng, Nt, Nx, Ny, Nz, Nmu, N_azi)],
        ["total", (Ns, Ng, Nt, Nx, Ny, Nz, Nmu, N_azi)],
        ["current", (Ns, Ng, Nt, Nx, Ny, Nz, 3)],
        ["eddington", (Ns, Ng, Nt, Nx, Ny, Nz, 6)],
        ["exit", (Ns, Ng, Nt, 2, Ny, Nz, Nmu, N_azi)],
    ]

    # Add score flags to structure
    for i in range(len(scores_shapes)):
        name = scores_shapes[i][0]
        struct += [(name, bool_)]

    # Add scores to structure
    scores_struct = []
    for i in range(len(scores_shapes)):
        name = scores_shapes[i][0]
        shape = scores_shapes[i][1]
        if not tally_card[name]:
            shape = (0,) * len(shape)
        scores_struct += [(name, make_type_uq_score(shape))]
    scores = np.dtype(scores_struct)
    struct += [("score", scores)]

    # Make tally structure
    uq_tally = np.dtype(struct)


def make_type_uq(input_deck):
    global uq, uq_nuc, uq_mat

    #    def make_type_parameter(shape):
    #        return np.dtype(
    #            [
    #                ("tag", str_),             # nuclides, materials, surfaces, sources
    #                ("ID", int64),
    #                ("key", str_),
    #                ("mean", float64, shape),
    #                ("delta", float64, shape),
    #                ("distribution", str_),
    #                ("rng_seed", uint64),
    #            ]
    #        )

    def make_type_parameter(G, J, decay=False):
        # Fields are things that can have deltas
        struct = [
            ("speed", float64, (G,)),
            ("capture", float64, (G,)),
            ("scatter", float64, (G, G)),
            ("fission", float64, (G,)),
            ("nu_s", float64, (G,)),
            ("nu_p", float64, (G,)),
            ("nu_d", float64, (G, J)),
            ("chi_p", float64, (G, G)),
        ]
        struct += [("decay", float64, (J,)), ("chi_d", float64, (J, G))]
        return np.dtype(struct)

    # Size numbers
    G = input_deck.materials[0]["G"]
    J = input_deck.materials[0]["J"]

    # UQ deck
    uq_deck = input_deck.uq_deltas

    uq_nuc = make_type_parameter(G, J, True)
    uq_mat = make_type_parameter(G, J)

    flags = np.dtype(
        [
            ("speed", bool_),
            ("decay", bool_),
            ("total", bool_),
            ("capture", bool_),
            ("scatter", bool_),
            ("fission", bool_),
            ("nu_s", bool_),
            ("nu_f", bool_),
            ("nu_p", bool_),
            ("nu_d", bool_),
            ("chi_s", bool_),
            ("chi_p", bool_),
            ("chi_d", bool_),
        ]
    )
    info = np.dtype([("distribution", str_), ("ID", int64), ("rng_seed", uint64)])

    container = np.dtype(
        [("mean", uq_nuc), ("delta", uq_mat), ("flags", flags), ("info", info)]
    )

    N_nuclide = len(uq_deck["nuclides"])
    N_material = len(uq_deck["materials"])
    uq = np.dtype(
        [("nuclides", container, (N_nuclide,)), ("materials", container, (N_material,))]
    )


param_names = ["tag", "ID", "key", "mean", "delta", "distribution", "rng_seed"]


# ==============================================================================
# Global
# ==============================================================================


def make_type_global(input_deck):
    global global_

    # Get modes
    mode_CE = input_deck.setting["mode_CE"]
    mode_MG = input_deck.setting["mode_MG"]

    # Numbers of objects
    N_nuclide = len(input_deck.nuclides)
    N_material = len(input_deck.materials)
    N_surface = len(input_deck.surfaces)
    N_cell = len(input_deck.cells)
    N_source = len(input_deck.sources)
    N_universe = len(input_deck.universes)
    N_lattice = len(input_deck.lattices)

    # Simulation parameters
    N_particle = input_deck.setting["N_particle"]
    N_precursor = input_deck.setting["N_precursor"]
    N_cycle = input_deck.setting["N_cycle"]

    # Particle bank buffers
    bank_active_buff = input_deck.setting["bank_active_buff"]
    bank_census_buff = input_deck.setting["bank_census_buff"]

    # Number of precursor groups
    if mode_MG:
        J = input_deck.materials[0]["J"]
    if mode_CE:
        J = 6

    # Number of work
    N_work = math.ceil(N_particle / MPI.COMM_WORLD.Get_size())
    N_work_precursor = math.ceil(N_precursor / MPI.COMM_WORLD.Get_size())

    # Particle bank types
    bank_active = particle_bank(1 + bank_active_buff)
    if input_deck.setting["mode_eigenvalue"] or input_deck.setting["N_census"] > 1:
        print(int((1 + bank_census_buff) * N_work))
        bank_census = particle_bank(int((1 + bank_census_buff) * N_work))
        bank_source = particle_bank(int((1 + bank_census_buff) * N_work))
    else:
        bank_census = particle_bank(0)
        bank_source = particle_bank(0)
    bank_precursor = precursor_bank(0)

    # Particle tracker
    N_track = 0
    if input_deck.setting["track_particle"]:
        N_track = N_work * 1000

    # iQMC bank adjustment
    if input_deck.technique["iQMC"]:
        bank_source = particle_bank(N_work)
        if input_deck.setting["mode_eigenvalue"]:
            bank_census = particle_bank(0)

    # Source and IC files bank adjustments
    if not input_deck.setting["mode_eigenvalue"]:
        if input_deck.setting["source_file"]:
            bank_source = particle_bank(N_work)
        if input_deck.setting["IC_file"]:
            bank_source = particle_bank(N_work)
            bank_precursor = precursor_bank(N_precursor)

    if (
        input_deck.setting["source_file"] and not input_deck.setting["mode_eigenvalue"]
    ) or input_deck.technique["iQMC"]:
        bank_source = particle_bank(N_work)

    # GLobal type
    global_ = np.dtype(
        [
            ("nuclides", nuclide, (N_nuclide,)),
            ("materials", material, (N_material,)),
            ("surfaces", surface, (N_surface,)),
            ("cells", cell, (N_cell,)),
            ("universes", universe, (N_universe,)),
            ("lattices", lattice, (N_lattice,)),
            ("sources", source, (N_source,)),
            ("tally", tally),
            ("setting", setting),
            ("technique", technique),
            ("bank_active", bank_active),
            ("bank_census", bank_census),
            ("bank_source", bank_source),
            ("bank_precursor", bank_precursor),
            ("k_eff", float64),
            ("k_cycle", float64, (N_cycle,)),
            ("k_avg", float64),
            ("k_sdv", float64),
            ("n_avg", float64),  # Neutron density
            ("n_sdv", float64),
            ("n_max", float64),
            ("C_avg", float64),  # Precursor density
            ("C_sdv", float64),
            ("C_max", float64),
            ("k_avg_running", float64),
            ("k_sdv_running", float64),
            ("gyration_radius", float64, (N_cycle,)),
            ("idx_cycle", int64),
            ("cycle_active", bool_),
            ("eigenvalue_tally_nuSigmaF", float64),
            ("eigenvalue_tally_n", float64),
            ("eigenvalue_tally_C", float64),
            ("idx_census", int64),
            ("idx_batch", int64),
            ("mpi_size", int64),
            ("mpi_rank", int64),
            ("mpi_master", bool_),
            ("mpi_work_start", int64),
            ("mpi_work_size", int64),
            ("mpi_work_size_total", int64),
            ("mpi_work_start_precursor", int64),
            ("mpi_work_size_precursor", int64),
            ("mpi_work_size_total_precursor", int64),
            ("runtime_total", float64),
            ("runtime_preparation", float64),
            ("runtime_simulation", float64),
            ("runtime_output", float64),
            ("runtime_bank_management", float64),
            ("particle_track", float64, (N_track, 8)),
            ("particle_track_N", int64),
            ("particle_track_history_ID", int64),
            ("particle_track_particle_ID", int64),
            ("precursor_strength", float64),
        ]
    )


# ==============================================================================
# Util
# ==============================================================================


def make_type_mesh(card):
    Nx = len(card["x"]) - 1
    Ny = len(card["y"]) - 1
    Nz = len(card["z"]) - 1
    Nt = len(card["t"]) - 1
    Nmu = len(card["mu"]) - 1
    N_azi = len(card["azi"]) - 1
    Ng = len(card["g"]) - 1
    return (
        np.dtype(
            [
                ("x", float64, (Nx + 1,)),
                ("y", float64, (Ny + 1,)),
                ("z", float64, (Nz + 1,)),
                ("t", float64, (Nt + 1,)),
                ("mu", float64, (Nmu + 1,)),
                ("azi", float64, (N_azi + 1,)),
                ("g", float64, (Ng + 1,)),
            ]
        ),
        Nx,
        Ny,
        Nz,
        Nt,
        Nmu,
        N_azi,
        Ng,
    )


def make_type_mesh_(card):
    Nx = len(card["x"]) - 1
    Ny = len(card["y"]) - 1
    Nz = len(card["z"]) - 1
    Nt = len(card["t"]) - 1
    Nmu = len(card["mu"]) - 1
    N_azi = len(card["azi"]) - 1
    return (
        np.dtype(
            [
                ("x", float64, (Nx + 1,)),
                ("y", float64, (Ny + 1,)),
                ("z", float64, (Nz + 1,)),
                ("t", float64, (Nt + 1,)),
                ("mu", float64, (Nmu + 1,)),
                ("azi", float64, (N_azi + 1,)),
            ]
        ),
        Nx,
        Ny,
        Nz,
        Nt,
        Nmu,
        N_azi,
    )


mesh_names = ["x", "y", "z", "t", "mu", "azi", "g"]



"""
This module contains functions for setting MC/DC input deck.
The input deck class is defined in `card.py` and instantiated in `global_.py`.
"""

import h5py, math, mpi4py, os
import numpy as np
import scipy as sp

import mcdc.type_ as type_

from mcdc.card import (
    SurfaceHandle,
    make_card_nuclide,
    make_card_material,
    make_card_surface,
    make_card_cell,
    make_card_universe,
    make_card_lattice,
    make_card_source,
    make_card_uq,
)
from mcdc.constant import (
    GYRATION_RADIUS_ALL,
    GYRATION_RADIUS_INFINITE_X,
    GYRATION_RADIUS_INFINITE_Y,
    GYRATION_RADIUS_INFINITE_Z,
    GYRATION_RADIUS_ONLY_X,
    GYRATION_RADIUS_ONLY_Y,
    GYRATION_RADIUS_ONLY_Z,
    PCT_NONE,
    PCT_COMBING,
    PCT_COMBING_WEIGHT,
    INF,
    PI,
    SHIFT,
)
from mcdc.print_ import print_error

# Get and rename mcdc global variables
import mcdc.global_ as mcdc


def nuclide(
    capture=None,
    scatter=None,
    fission=None,
    nu_s=None,
    nu_p=None,
    nu_d=None,
    chi_p=None,
    chi_d=None,
    speed=None,
    decay=None,
    sensitivity=False,
    dsm_Np=1.0,
):
    """
    Create a nuclide card.

    Parameters
    ----------
    capture : numpy.ndarray (1D), optional
        Capture microscopic cross-section [barn]
    scatter : numpy.ndarray (2D), optional
        Differential scattering microscopic cross-section [gout, gin] [barn].
    fission : numpy.ndarray (1D), optional
        Fission microscopic cross-section [barn].
    nu_s : numpy.ndarray (1D), optional
        Scattering multiplication.
    nu_p : numpy.ndarray (1D), optional
        Prompt fission neutron yield.
    nu_d : numpy.ndarray (2D), optional
        Delayed neutron precursor yield [dg, gin].
    chi_p : numpy.ndarray (2D), optional
        Prompt fission spectrum [gout, gin]
    chi_d : numpy.ndarray (2D), optional
        Delayed neutron spectrum [gout, dg]
    speed : numpy.ndarray (1D), optional
        Energy group speed [cm/s]
    decay : numpy.ndarray (1D), optional
        Precursor group decay constant [/s]
    sensitivity : bool, optional
        Set to `True` to calculate sensitivities to the nuclide
    dsm_Np : float
        Average number of derivative particles produced at each
        sensitivity nuclide collision

    Returns
    -------
    dictionary
        A nuclide card

    Notes
    -----
    Parameters are set to zeros by default. Energy group size G is determined by the
    size of `capture`, `scatter`, or `fission`. Thus, at least `capture`, `scatter`,
    or `fission` needs to be provided. `nu_p` or `nu_d` is needed if `fission` is
    provided. `chi_p` and `chi_d` are needed if `nu_p` and `nu_d` are provided,
    respectively, and G > 1. Delayed neutron precursor group size J is determined by
    the size of `nu_d`; if `nu_d` is not given, J = 0.

    See also
    --------
    mcdc.material : A material can be defined as a collection of nuclides.
    """
    # Energy group size
    if capture is not None:
        G = len(capture)
    elif scatter is not None:
        G = len(scatter)
    elif fission is not None:
        G = len(fission)
    else:
        print_error("Need to supply capture, scatter, or fission to mcdc.nuclide")

    # Delayed group size
    J = 0
    if nu_d is not None:
        J = len(nu_d)

    # Make nuclide card
    card = make_card_nuclide(G, J)
    card["ID"] = len(mcdc.input_deck.nuclides)

    # Speed (vector of size G)
    if speed is not None:
        card["speed"][:] = speed[:]

    # Decay constant (vector of size J)
    if decay is not None:
        card["decay"][:] = decay[:]

    # Cross-sections (vector of size G)
    if capture is not None:
        card["capture"][:] = capture[:]
    if scatter is not None:
        card["scatter"][:] = np.sum(scatter, 0)[:]
    if fission is not None:
        card["fission"][:] = fission[:]
        card["fissionable"] = True
    card["total"][:] = card["capture"] + card["scatter"] + card["fission"]

    # Scattering multiplication (vector of size G)
    if nu_s is not None:
        card["nu_s"][:] = nu_s[:]

    # Check if nu_p or nu_d is not provided, give fission
    if fission is not None:
        if nu_p is None and nu_d is None:
            print_error("Need to supply nu_p or nu_d for fissionable mcdc.nuclide")

    # Prompt fission production (vector of size G)
    if nu_p is not None:
        card["nu_p"][:] = nu_p[:]

    # Delayed fission production (matrix of size GxJ)
    if nu_d is not None:
        # Transpose: [dg, gin] -> [gin, dg]
        card["nu_d"][:, :] = np.swapaxes(nu_d, 0, 1)[:, :]

    # Total fission production (vector of size G)
    card["nu_f"] += card["nu_p"]
    for j in range(J):
        card["nu_f"] += card["nu_d"][:, j]

    # Scattering spectrum (matrix of size GxG)
    if scatter is not None:
        # Transpose: [gout, gin] -> [gin, gout]
        card["chi_s"][:, :] = np.swapaxes(scatter, 0, 1)[:, :]
        for g in range(G):
            if card["scatter"][g] > 0.0:
                card["chi_s"][g, :] /= card["scatter"][g]

    # Prompt fission spectrum (matrix of size GxG)
    if nu_p is not None:
        if G == 1:
            card["chi_p"][:, :] = np.array([[1.0]])
        elif chi_p is None:
            print_error("Need to supply chi_p if nu_p is provided and G > 1")
        else:
            # Transpose: [gout, gin] -> [gin, gout]
            card["chi_p"][:, :] = np.swapaxes(chi_p, 0, 1)[:, :]
            # Normalize
            for g in range(G):
                if np.sum(card["chi_p"][g, :]) > 0.0:
                    card["chi_p"][g, :] /= np.sum(card["chi_p"][g, :])

    # Delayed fission spectrum (matrix of size JxG)
    if nu_d is not None:
        if G == 1:
            card["chi_d"][:, :] = np.ones([J, G])
        else:
            if chi_d is None:
                print_error("Need to supply chi_d if nu_d is provided  and G > 1")
            # Transpose: [gout, dg] -> [dg, gout]
            card["chi_d"][:, :] = np.swapaxes(chi_d, 0, 1)[:, :]
        # Normalize
        for dg in range(J):
            if np.sum(card["chi_d"][dg, :]) > 0.0:
                card["chi_d"][dg, :] /= np.sum(card["chi_d"][dg, :])

    # Sensitivity setup
    if sensitivity:
        # Set flag
        card["sensitivity"] = True
        mcdc.input_deck.technique["sensitivity"] = True
        mcdc.input_deck.technique["weighted_emission"] = False

        # Set ID
        mcdc.input_deck.setting["N_sensitivity"] += 1
        card["sensitivity_ID"] = mcdc.input_deck.setting["N_sensitivity"]

        # Set dsm_Np
        card["dsm_Np"] = dsm_Np

    # Push card
    mcdc.input_deck.nuclides.append(card)
    return card


def material(
    nuclides=None,
    capture=None,
    scatter=None,
    fission=None,
    nu_s=None,
    nu_p=None,
    nu_d=None,
    chi_p=None,
    chi_d=None,
    speed=None,
    decay=None,
    name="P",
    sensitivity=False,
    dsm_Np=1.0,
):
    """
    Create a material card.

    The material card is defined either as a collection of nuclides or directly by its
    macroscopic constants.

    Parameters
    ----------
    nuclides : list of tuple of (dictionary, float), optional
        List of pairs of nuclide card and its density [/barn-cm]
    capture : numpy.ndarray (1D), optional
        Capture macroscopic cross-section [/cm]
    scatter : numpy.ndarray (2D), optional
        Differential scattering macroscopic cross-section [gout, gin] [/cm].
    fission : numpy.ndarray (1D), optional
        Fission macroscopic cross-section [/cm].
    nu_s : numpy.ndarray (1D), optional
        Scattering multiplication.
    nu_p : numpy.ndarray (1D), optional
        Prompt fission neutron yield.
    nu_d : numpy.ndarray (2D), optional
        Delayed neutron precursor yield [dg, gin].
    chi_p : numpy.ndarray (2D), optional
        Prompt fission spectrum [gout, gin]
    chi_d : numpy.ndarray (2D), optional
        Delayed neutron spectrum [gout, dg]
    speed : numpy.ndarray (1D), optional
        Energy group speed [cm/s]
    decay : numpy.ndarray (1D), optional
        Precursor group decay constant [/s]
    sensitivity : bool, optional
        Set to `True` to calculate sensitivities to the material
        (only relevant for single-nuclide material)
    dsm_Np : float
        Average number of derivative particles produced at each
        sensitivity material collision (only relevant for single_nuclide material)

    Returns
    -------
    dictionary
        A material card

    See also
    --------
    mcdc.nuclide : A material can be defined as a collection of nuclides.
    """
    # If nuclides are not given, and macroscopic constants are given instead,
    # create a nuclide card and set a single-nuclide material
    if nuclides is None:
        card_nuclide = nuclide(
            capture,
            scatter,
            fission,
            nu_s,
            nu_p,
            nu_d,
            chi_p,
            chi_d,
            speed,
            decay,
            sensitivity,
            dsm_Np,
        )
        nuclides = [[card_nuclide, 1.0]]

    # Number of nuclides
    N_nuclide = len(nuclides)

    # Continuous energy mode?
    if isinstance(nuclides[0][0], str):
        mcdc.input_deck.setting["mode_CE"] = True
        mcdc.input_deck.setting["mode_MG"] = False

        # Make material card
        card = make_card_material(N_nuclide)
        card["ID"] = len(mcdc.input_deck.materials)

        # Set the nuclides
        for i in range(N_nuclide):
            nuc_name = nuclides[i][0]
            density = nuclides[i][1]
            if not nuclide_registered(nuc_name):
                nuc_ID = len(mcdc.input_deck.nuclides)
                nuc_card = make_card_nuclide()
                nuc_card["name"] = nuc_name
                nuc_card["ID"] = nuc_ID

                dir_name = os.getenv("MCDC_XSLIB")
                with h5py.File(dir_name + "/" + nuc_name + ".h5", "r") as f:
                    if max(f["fission"][:]) > 0.0:
                        nuc_card["fissionable"] = True

                mcdc.input_deck.nuclides.append(nuc_card)
            else:
                nuc_card = get_nuclide(nuc_name)
            card["nuclide_IDs"][i] = nuc_card["ID"]
            card["nuclide_densities"][i] = density

        # Add to deck
        mcdc.input_deck.materials.append(card)
        return card

    # Nuclide and group sizes
    G = nuclides[0][0]["G"]
    J = nuclides[0][0]["J"]

    # Make material card
    card = make_card_material(N_nuclide, G, J)
    card["ID"] = len(mcdc.input_deck.materials)

    if name is not None:
        card["name"] = name
    else:
        card["name"] = card["ID"]

    # Calculate basic XS and determine sensitivity flag
    for i in range(N_nuclide):
        nuc = nuclides[i][0]
        density = nuclides[i][1]
        card["nuclide_IDs"][i] = nuc["ID"]
        card["nuclide_densities"][i] = density
        for tag in ["capture", "scatter", "fission", "total", "sensitivity"]:
            card[tag] += nuc[tag] * density
    card["sensitivity"] = bool(card["sensitivity"])

    # Calculate effective speed
    # Current approach: weighted by nuclide macroscopic total cross section
    # TODO: other more appropriate way?
    for i in range(N_nuclide):
        nuc = nuclides[i][0]
        density = nuclides[i][1]
        card["speed"] += nuc["speed"] * nuc["total"] * density
    # Check if vacuum material
    if max(card["total"]) == 0.0:
        card["speed"][:] = nuc["speed"][:]
    else:
        card["speed"] /= card["total"]

    # Calculate effective spectra and multiplicities of scattering and prompt fission
    if max(card["scatter"]) > 0.0:
        nuSigmaS = np.zeros((G, G), dtype=float)
        for i in range(N_nuclide):
            nuc = nuclides[i][0]
            density = nuclides[i][1]
            SigmaS = np.diag(nuc["scatter"]) * density
            nu_s = np.diag(nuc["nu_s"])
            chi_s = np.transpose(nuc["chi_s"])
            nuSigmaS += chi_s.dot(nu_s.dot(SigmaS))
        chi_nu_s = nuSigmaS.dot(np.diag(1.0 / card["scatter"]))
        card["nu_s"] = np.sum(chi_nu_s, axis=0)
        card["chi_s"] = np.transpose(chi_nu_s.dot(np.diag(1.0 / card["nu_s"])))
    if max(card["fission"]) > 0.0:
        nuSigmaF = np.zeros((G, G), dtype=float)
        for i in range(N_nuclide):
            nuc = nuclides[i][0]
            density = nuclides[i][1]
            SigmaF = np.diag(nuc["fission"]) * density
            nu_p = np.diag(nuc["nu_p"])
            chi_p = np.transpose(nuc["chi_p"])
            nuSigmaF += chi_p.dot(nu_p.dot(SigmaF))
        chi_nu_p = nuSigmaF.dot(np.diag(1.0 / card["fission"]))
        card["nu_p"] = np.sum(chi_nu_p, axis=0)
        card["chi_p"] = np.transpose(chi_nu_p.dot(np.diag(1.0 / card["nu_p"])))

    # Calculate delayed and total fission multiplicities
    if max(card["fission"]) > 0.0:
        card["nu_f"][:] = card["nu_p"][:]
        for j in range(J):
            total = np.zeros(G)
            for i in range(N_nuclide):
                nuc = nuclides[i][0]
                density = nuclides[i][1]
                total += nuc["nu_d"][:, j] * nuc["fission"] * density
            card["nu_d"][:, j] = total / card["fission"]
            card["nu_f"] += card["nu_d"][:, j]

    # Push card
    mcdc.input_deck.materials.append(card)
    return card


def surface(type_, bc="interface", sensitivity=False, dsm_Np=1.0, **kw):
    """
    Create a surface card and return SurfaceHandle to define cell domain.

    Parameters
    ----------
    type_ : {'plane-x', 'plane-y', 'plane-z', 'plane', 'cylinder-x', 'cylinder-y',
             'cylinder-z', 'sphere', 'quadric'}
        Surface type.
    bc : {'interface', 'vacuum', 'reflective'}
        Surface boundary condition.
    sensitivity : bool, optional
        Set to `True` to calculate sensitivities to the nuclide
    dsm_Np : int
        Average number of derivative particles produced at each
        sensitivity surface crossing

    Other Parameters
    ----------------
    x : {float, array_like}
        x-position [cm] for `plane-x`. If a vector is passed, positions of the surface
        at the times specified by the parameter `t`.
    y : {float, array_like}
        y-position [cm] for `plane-y`. If a vector is passed, positions of the surface
        at the times specified by the parameter `t`.
    z : {float, array_like}
        z-position [cm] for `plane-z`. If a vector is passed, positions of the surface
        at the times specified by the parameter `t`.
    center : array_like
        Center point [cm] for `cylinder-x` (y,z), `cylinder-y` (x,z),
        `cylinder-z` (x,y), and `sphere` (x,y,z).
    radius : float
        Radius [cm] for `cylinder-x`, `cylinder-y`, `cylinder-z`, and `sphere`.
    A, B, C, D : float
        Coefficients [cm] for `plane`.
    A, B, C, D, E, F, G, H, I, J : float
        Coefficients [cm] for `quadric`.

    Returns
    -------
    SurfaceHandle
        A surface handle used for assigning surface, and its sense, to a cell card.

    See also
    --------
    mcdc.cell : SurfaceHandle is used to define cell domain
    """
    # Make surface card
    card = make_card_surface()
    card["ID"] = len(mcdc.input_deck.surfaces)

    # Check if the selected type is supported
    type_ = check_support(
        "surface type",
        type_,
        [
            "plane-x",
            "plane-y",
            "plane-z",
            "plane",
            "cylinder-x",
            "cylinder-y",
            "cylinder-z",
            "sphere",
            "quadric",
        ],
    )

    # Boundary condition
    bc = check_support(
        "surface boundary condition",
        bc,
        [
            "interface",
            "vacuum",
            "reflective",
        ],
    )
    # Set bc flags
    if bc == "vacuum":
        card["vacuum"] = True
    elif bc == "reflective":
        card["reflective"] = True

    # Sensitivity
    if sensitivity:
        # Set flag
        card["sensitivity"] = True
        mcdc.input_deck.technique["sensitivity"] = True
        mcdc.input_deck.technique["weighted_emission"] = False

        # Set ID
        mcdc.input_deck.setting["N_sensitivity"] += 1
        card["sensitivity_ID"] = mcdc.input_deck.setting["N_sensitivity"]

        # Set dsm_Np
        card["dsm_Np"] = dsm_Np

    # ==========================================================================
    # Surface attributes
    # ==========================================================================
    # Axx + Byy + Czz + Dxy + Exz + Fyz + Gx + Hy + Iz + J(t) = 0
    #   J(t) = J0_i + J1_i*t for t in [t_{i-1}, t_i), t_0 = 0

    card["type"] = type_

    # Set up surface attributes
    if type_ == "plane-x":
        check_requirement("surface plane-x", kw, ["x"])
        card["G"] = 1.0
        card["linear"] = True
        if type(kw.get("x")) in [type([]), type(np.array([]))]:
            _set_J(kw.get("x"), kw.get("t"), card)
        else:
            card["J"][0, 0] = -kw.get("x")
    elif type_ == "plane-y":
        check_requirement("surface plane-y", kw, ["y"])
        card["H"] = 1.0
        card["linear"] = True
        if type(kw.get("y")) in [type([]), type(np.array([]))]:
            _set_J(kw.get("y"), kw.get("t"), card)
        else:
            card["J"][0, 0] = -kw.get("y")
    elif type_ == "plane-z":
        check_requirement("surface plane-z", kw, ["z"])
        card["I"] = 1.0
        card["linear"] = True
        if type(kw.get("z")) in [type([]), type(np.array([]))]:
            _set_J(kw.get("z"), kw.get("t"), card)
        else:
            card["J"][0, 0] = -kw.get("z")
    elif type_ == "plane":
        check_requirement("surface plane", kw, ["A", "B", "C", "D"])
        card["G"] = kw.get("A")
        card["H"] = kw.get("B")
        card["I"] = kw.get("C")
        card["J"][0, 0] = kw.get("D")
        card["linear"] = True
    elif type_ == "cylinder-x":
        check_requirement("surface cylinder-x", kw, ["center", "radius"])
        y, z = kw.get("center")[:]
        r = kw.get("radius")
        card["B"] = 1.0
        card["C"] = 1.0
        card["H"] = -2.0 * y
        card["I"] = -2.0 * z
        card["J"][0, 0] = y**2 + z**2 - r**2
    elif type_ == "cylinder-y":
        check_requirement("surface cylinder-y", kw, ["center", "radius"])
        x, z = kw.get("center")[:]
        r = kw.get("radius")
        card["A"] = 1.0
        card["C"] = 1.0
        card["G"] = -2.0 * x
        card["I"] = -2.0 * z
        card["J"][0, 0] = x**2 + z**2 - r**2
    elif type_ == "cylinder-z":
        check_requirement("surface cylinder-z", kw, ["center", "radius"])
        x, y = kw.get("center")[:]
        r = kw.get("radius")
        card["A"] = 1.0
        card["B"] = 1.0
        card["G"] = -2.0 * x
        card["H"] = -2.0 * y
        card["J"][0, 0] = x**2 + y**2 - r**2
    elif type_ == "sphere":
        check_requirement("surface sphere", kw, ["center", "radius"])
        x, y, z = kw.get("center")[:]
        r = kw.get("radius")
        card["A"] = 1.0
        card["B"] = 1.0
        card["C"] = 1.0
        card["G"] = -2.0 * x
        card["H"] = -2.0 * y
        card["I"] = -2.0 * z
        card["J"][0, 0] = x**2 + y**2 + z**2 - r**2
    elif type_ == "quadric":
        check_requirement(
            "surface quadric", kw, ["A", "B", "C", "D", "E", "F", "G", "H", "I", "J"]
        )
        card["A"] = kw.get("A")
        card["B"] = kw.get("B")
        card["C"] = kw.get("C")
        card["D"] = kw.get("D")
        card["E"] = kw.get("E")
        card["F"] = kw.get("F")
        card["G"] = kw.get("G")
        card["H"] = kw.get("H")
        card["I"] = kw.get("I")
        card["J"][0, 0] = kw.get("J")

    # Set normal vector if linear
    if card["linear"]:
        nx = card["G"]
        ny = card["H"]
        nz = card["I"]
        # Normalize
        norm = (nx**2 + ny**2 + nz**2) ** 0.5
        card["nx"] = nx / norm
        card["ny"] = ny / norm
        card["nz"] = nz / norm

    # Push card
    mcdc.input_deck.surfaces.append(card)
    return SurfaceHandle(card)


def _set_J(x, t, card):
    # Edit and add the edges
    t[0] = -SHIFT
    t = np.append(t, INF)
    x = np.append(x, x[-1])

    # Reset the constants
    card["J"] = np.zeros([0, 2])
    card["t"] = np.array([-SHIFT])

    # Iterate over inputs
    idx = 0
    for i in range(len(t) - 1):
        # Skip if step
        if t[i] == t[i + 1]:
            continue

        # Calculate constants
        J0 = x[i]
        J1 = (x[i + 1] - x[i]) / (t[i + 1] - t[i])

        # Append to card
        card["J"] = np.append(card["J"], [[J0, J1]], axis=0)
        card["t"] = np.append(card["t"], t[i + 1])

    card["J"] *= -1
    card["N_slice"] = len(card["J"])


def cell(surfaces_flags, fill, lattice_center=None):
    N_surface = len(surfaces_flags)

    # Make cell card
    card = make_card_cell(N_surface)
    card["ID"] = len(mcdc.input_deck.cells)

    # Surfaces and flags
    for i in range(N_surface):
        card["surface_IDs"][i] = surfaces_flags[i][0]["ID"]
        card["positive_flags"][i] = surfaces_flags[i][1]

    # Lattice cell?
    if fill["tag"] == "Lattice":
        card["lattice"] = True
        card["lattice_ID"] = fill["ID"]
        if lattice_center is not None:
            card["lattice_center"] = np.array(lattice_center)

    # Material cell
    else:
        card["material_ID"] = fill["ID"]
        card["material_name"] = fill["name"]

    # Push card
    mcdc.input_deck.cells.append(card)
    return card


def universe(cells, root=False):
    N_cell = len(cells)

    # Set default card values (c.f. type_.py)
    if not root:
        card = make_card_universe(N_cell)
        card["ID"] = len(mcdc.input_deck.universes)
    else:
        card = mcdc.input_deck.universes[0]
        card["N_cell"] = N_cell
        card["cell_IDs"] = np.zeros(N_cell, dtype=int)

    # Cells
    for i in range(N_cell):
        card["cell_IDs"][i] = cells[i]["ID"]

    # Push card
    if not root:
        mcdc.input_deck.universes.append(card)

    return card


def lattice(x=None, y=None, z=None, universes=None):
    # Make lattice card
    card = make_card_lattice()
    card["ID"] = len(mcdc.input_deck.lattices)

    # Set mesh
    if x is not None:
        card["mesh"]["x0"] = x[0]
        card["mesh"]["dx"] = x[1]
        card["mesh"]["Nx"] = x[2]
    if y is not None:
        card["mesh"]["y0"] = y[0]
        card["mesh"]["dy"] = y[1]
        card["mesh"]["Ny"] = y[2]
    if z is not None:
        card["mesh"]["z0"] = z[0]
        card["mesh"]["dz"] = z[1]
        card["mesh"]["Nz"] = z[2]

    # Set universe IDs
    universe_IDs = np.array(universes, dtype=np.int64)
    ax_expand = []
    if x is None:
        ax_expand.append(2)
    if y is None:
        ax_expand.append(1)
    if z is None:
        ax_expand.append(0)
    for ax in ax_expand:
        universe_IDs = np.expand_dims(universe_IDs, axis=ax)

    # Change indexing structure: [z(flip), y(flip), x] --> [x, y, z]
    tmp = np.transpose(universe_IDs)
    tmp = np.flip(tmp, axis=1)
    card["universe_IDs"] = np.flip(tmp, axis=2)

    # Push card
    mcdc.input_deck.lattices.append(card)
    return card


def source(**kw):
    """
    Create a source card.

    Other Parameters
    ----------------
    point : array_like
        [x, y, z] point position for point source.
    x : array_like
        [x_min and x_max] for uniform source.
    y : array_like
        [y_min and y_max] for uniform source.
    z : array_like
        [z_min and z_max] for uniform source.
    isotropic : bool
        Flag for isotropic source
    direction : array_like
        [ux, uy, uz] unit vector for parallel beam source.
    white_direction : array_like
        [nx, ny, nz] unit vector of the normal outward direction of the surface
        at which isotropic surface source is emitted. Note that it is similar to the
        mechanics of the typical white boundary condition in reactor physics.
    energy : array_like
        [MG] Probability mass function of the energy group for multigroup source.
        [CE] 2D array of piecewise linear pdf [eV, value].
    time : array_like
        [t_min and t_max] in/at which source is emitted.
    prob : float
        Relative probability (or strength) of the source.

    Returns
    -------
    dictionary
        A source card
    """
    # Check the suplied keyword arguments
    for key in kw.keys():
        check_support(
            "source parameter",
            key,
            [
                "point",
                "x",
                "y",
                "z",
                "isotropic",
                "direction",
                "white_direction",
                "energy",
                "time",
                "prob",
            ],
            False,
        )

    # Get keyword arguments
    point = kw.get("point")
    x = kw.get("x")
    y = kw.get("y")
    z = kw.get("z")
    isotropic = kw.get("isotropic")
    direction = kw.get("direction")
    white = kw.get("white_direction")
    energy = kw.get("energy")
    time = kw.get("time")
    prob = kw.get("prob")

    # Make source card
    card = make_card_source()
    card["ID"] = len(mcdc.input_deck.sources)

    # Set position
    if point is not None:
        card["x"] = point[0]
        card["y"] = point[1]
        card["z"] = point[2]
    else:
        card["box"] = True
        if x is not None:
            card["box_x"] = np.array(x)
        if y is not None:
            card["box_y"] = np.array(y)
        if z is not None:
            card["box_z"] = np.array(z)

    # Set direction
    if white is not None:
        card["isotropic"] = False
        card["white"] = True
        ux = white[0]
        uy = white[1]
        uz = white[2]
        # Normalize
        norm = (ux**2 + uy**2 + uz**2) ** 0.5
        card["white_x"] = ux / norm
        card["white_y"] = uy / norm
        card["white_z"] = uz / norm
    elif direction is not None:
        card["isotropic"] = False
        ux = direction[0]
        uy = direction[1]
        uz = direction[2]
        # Normalize
        norm = (ux**2 + uy**2 + uz**2) ** 0.5
        card["ux"] = ux / norm
        card["uy"] = uy / norm
        card["uz"] = uz / norm

    # Set energy
    if energy is not None:
        if mcdc.input_deck.setting["mode_MG"]:
            group = np.array(energy)
            # Normalize
            card["group"] = group / np.sum(group)
        if mcdc.input_deck.setting["mode_CE"]:
            energy = np.array(energy)
            # Resize
            card["energy"] = np.zeros(energy.shape)
            # Set energy
            card["energy"][0, :] = energy[0, :]
            # Normalize pdf
            card["energy"][1, :] = energy[1, :] / np.trapz(energy[1, :], x=energy[0, :])
            # Make cdf
            card["energy"][1, :] = sp.integrate.cumulative_trapezoid(
                card["energy"][1], x=card["energy"][0], initial=0.0
            )

    # Set time
    if time is not None:
        card["time"] = np.array(time)

    # Set probability
    if prob is not None:
        card["prob"] = prob

    # Push card
    mcdc.input_deck.sources.append(card)
    return card


def tally(
    scores,
    x=np.array([-INF, INF]),
    y=np.array([-INF, INF]),
    z=np.array([-INF, INF]),
    t=np.array([-INF, INF]),
    mu=np.array([-1.0, 1.0]),
    azi=np.array([-PI, PI]),
    g=np.array([-INF, INF]),
    E=np.array([0.0, INF]),
):
    # Get tally card
    card = mcdc.input_deck.tally

    # Set mesh
    card["mesh"]["x"] = x
    card["mesh"]["y"] = y
    card["mesh"]["z"] = z
    card["mesh"]["t"] = t
    card["mesh"]["mu"] = mu
    card["mesh"]["azi"] = azi

    # Set energy group grid
    if type(g) == type("string") and g == "all":
        G = mcdc.input_deck.materials[0]["G"]
        card["mesh"]["g"] = np.linspace(0, G, G + 1) - 0.5
    else:
        card["mesh"]["g"] = g
    if mcdc.input_deck.setting["mode_CE"]:
        card["mesh"]["g"] = E

    # Set score flags
    for s in scores:
        found = False
        for score_name in type_.score_list:
            if s.replace("-", "_") == score_name:
                card["tracklength"] = True
                card[score_name] = True
                found = True
                break
        if not found:
            print_error("Unknown tally score %s" % s)

    return card


# ==============================================================================
# Setting
# ==============================================================================


def setting(**kw):
    # Check the suplied keyword arguments
    for key in kw.keys():
        check_support(
            "setting parameter",
            key,
            [
                "N_particle",
                "N_batch",
                "rng_seed",
                "time_boundary",
                "progress_bar",
                "output_name",
                "save_input_deck",
                "particle_tracker",
                "k_eff",
                "source_file",
                "IC_file",
                "active_bank_buff",
                "census_bank_buff",
            ],
            False,
        )

    # Get keyword arguments
    N_particle = kw.get("N_particle")
    N_batch = kw.get("N_batch")
    rng_seed = kw.get("rng_seed")
    time_boundary = kw.get("time_boundary")
    progress_bar = kw.get("progress_bar")
    output = kw.get("output_name")
    save_input_deck = kw.get("save_input_deck")
    particle_tracker = kw.get("particle_tracker")
    k_eff = kw.get("k_eff")
    source_file = kw.get("source_file")
    IC_file = kw.get("IC_file")
    bank_active_buff = kw.get("active_bank_buff")
    bank_census_buff = kw.get("census_bank_buff")

    # Check if setting card has been initialized
    card = mcdc.input_deck.setting

    # Number of particles
    if N_particle is not None:
        card["N_particle"] = int(N_particle)

    # Number of batches
    if N_batch is not None:
        card["N_batch"] = int(N_batch)

    # Time boundary
    if time_boundary is not None:
        card["time_boundary"] = time_boundary

    # RNG seed and stride
    if rng_seed is not None:
        card["rng_seed"] = rng_seed

    # Output .h5 file name
    if output is not None:
        card["output_name"] = output

    # Progress bar
    if progress_bar is not None:
        card["progress_bar"] = progress_bar

    # k effective
    if k_eff is not None:
        card["k_init"] = k_eff

    # Maximum active bank size
    if bank_active_buff is not None:
        card["bank_active_buff"] = int(bank_active_buff)

    # Census bank size multiplier
    if bank_census_buff is not None:
        card["bank_census_buff"] = int(bank_census_buff)

    # Particle tracker
    if particle_tracker is not None:
        card["track_particle"] = particle_tracker
        if particle_tracker and mpi4py.MPI.COMM_WORLD.Get_size() > 1:
            print_error("Particle tracker currently only runs on a single MPI rank")

    # Save input deck?
    if save_input_deck is not None:
        card["save_input_deck"] = save_input_deck

    # Source file
    if source_file is not None:
        card["source_file"] = True
        card["source_file_name"] = source_file

        # Set number of particles
        card_setting = mcdc.input_deck.setting
        with h5py.File(source_file, "r") as f:
            card_setting["N_particle"] = f["particles_size"][()]

    # IC file
    if IC_file is not None:
        card["IC_file"] = True
        card["IC_file_name"] = IC_file

        # Set number of particles
        card_setting = mcdc.input_deck.setting
        with h5py.File(IC_file, "r") as f:
            card_setting["N_particle"] = f["IC/neutrons_size"][()]
            card_setting["N_precursor"] = f["IC/precursors_size"][()]

    # TODO: Allow both source and IC files
    if IC_file and source_file:
        print_error("Using both source and IC files is not supported yet.")


def eigenmode(
    N_inactive=0, N_active=0, k_init=1.0, gyration_radius=None, save_particle=False
):
    # Update setting card
    card = mcdc.input_deck.setting
    card["N_inactive"] = N_inactive
    card["N_active"] = N_active
    card["N_cycle"] = N_inactive + N_active
    card["mode_eigenvalue"] = True
    card["k_init"] = k_init
    card["save_particle"] = save_particle

    # Gyration radius setup
    if gyration_radius is not None:
        card["gyration_radius"] = True
        if gyration_radius == "all":
            card["gyration_radius_type"] = GYRATION_RADIUS_ALL
        elif gyration_radius == "infinite-x":
            card["gyration_radius_type"] = GYRATION_RADIUS_INFINITE_X
        elif gyration_radius == "infinite-y":
            card["gyration_radius_type"] = GYRATION_RADIUS_INFINITE_Y
        elif gyration_radius == "infinite-z":
            card["gyration_radius_type"] = GYRATION_RADIUS_INFINITE_Z
        elif gyration_radius == "only-x":
            card["gyration_radius_type"] = GYRATION_RADIUS_ONLY_X
        elif gyration_radius == "only-y":
            card["gyration_radius_type"] = GYRATION_RADIUS_ONLY_Y
        elif gyration_radius == "only-z":
            card["gyration_radius_type"] = GYRATION_RADIUS_ONLY_Z
        else:
            print_error("Unknown gyration radius type")

    # Update tally card
    card = mcdc.input_deck.tally
    card["tracklength"] = True


# ==============================================================================
# Technique
# ==============================================================================


def implicit_capture():
    card = mcdc.input_deck.technique
    card["implicit_capture"] = True
    card["weighted_emission"] = False


def weighted_emission(flag):
    card = mcdc.input_deck.technique
    card["weighted_emission"] = flag


def population_control(pct="combing"):
    card = mcdc.input_deck.technique
    card["population_control"] = True
    card["weighted_emission"] = False
    if pct == "combing":
        card["pct"] = PCT_COMBING
    elif pct == "combing-weight":
        card["pct"] = PCT_COMBING_WEIGHT
    else:
        print_error("Unknown PCT type " + pct)


def branchless_collision():
    card = mcdc.input_deck.technique
    card["branchless_collision"] = True
    card["weighted_emission"] = False


def time_census(t):
    # Remove census beyond the final tally time grid point
    while True:
        if t[-1] >= mcdc.input_deck.tally["mesh"]["t"][-1]:
            t = t[:-1]
        else:
            break

    # Add the default, final census-at-infinity
    t = np.append(t, INF)

    # Set the time census parameters
    card = mcdc.input_deck.setting
    card["census_time"] = t
    card["N_census"] = len(t)


def weight_window(x=None, y=None, z=None, t=None, window=None, width=None):
    card = mcdc.input_deck.technique
    card["weight_window"] = True

    # Set width
    if width is not None:
        card["ww_width"] = width

    # Set mesh
    if x is not None:
        card["ww_mesh"]["x"] = x
    if y is not None:
        card["ww_mesh"]["y"] = y
    if z is not None:
        card["ww_mesh"]["z"] = z
    if t is not None:
        card["ww_mesh"]["t"] = t

    # Set window
    ax_expand = []
    if t is None:
        ax_expand.append(0)
    if x is None:
        ax_expand.append(1)
    if y is None:
        ax_expand.append(2)
    if z is None:
        ax_expand.append(3)
    window /= np.max(window)
    for ax in ax_expand:
        window = np.expand_dims(window, axis=ax)
    card["ww"] = window

    return card


def iQMC(
    g=None,
    t=None,
    x=None,
    y=None,
    z=None,
    phi0=None,
    source0=None,
    source_x0=None,
    source_y0=None,
    source_z0=None,
    source_xy0=None,
    source_xz0=None,
    source_yz0=None,
    source_xyz0=None,
    fission_source0=None,
    krylov_restart=None,
    fixed_source=None,
    scramble=False,
    maxitt=25,
    tol=1e-6,
    N_dim=6,
    seed=12345,
    preconditioner_sweeps=5,
    generator="halton",
    fixed_source_solver="source_iteration",
    eigenmode_solver="power_iteration",
    score=[],
):
    card = mcdc.input_deck.technique
    card["iQMC"] = True
    card["iqmc"]["tol"] = tol
    card["iqmc"]["maxitt"] = maxitt
    card["iqmc"]["generator"] = generator
    card["iqmc"]["N_dim"] = N_dim
    card["iqmc"]["scramble"] = scramble
    card["iqmc"]["seed"] = seed

    # Set mesh
    if g is not None:
        card["iqmc"]["mesh"]["g"] = g
    if t is not None:
        card["iqmc"]["mesh"]["t"] = t
    if x is not None:
        card["iqmc"]["mesh"]["x"] = x
    if y is not None:
        card["iqmc"]["mesh"]["y"] = y
    if z is not None:
        card["iqmc"]["mesh"]["z"] = z

    ax_expand = []
    if g is None:
        ax_expand.append(0)
    if t is None:
        ax_expand.append(1)
    if x is None:
        ax_expand.append(2)
    if y is None:
        ax_expand.append(3)
    if z is None:
        ax_expand.append(4)
    for ax in ax_expand:
        phi0 = np.expand_dims(phi0, axis=ax)
        if fixed_source is not None:
            fixed_source = np.expand_dims(fixed_source, axis=ax)
        else:
            fixed_source = np.zeros_like(phi0)

    if krylov_restart is None:
        krylov_restart = maxitt

    if source0 is None:
        source0 = np.zeros_like(phi0)

    if eigenmode_solver == "davidson":
        card["iqmc"]["krylov_vector_size"] += 1

    score_list = card["iqmc"]["score_list"]
    for name in score:
        score_list[name] = True

    if score_list["tilt-x"]:
        card["iqmc"]["krylov_vector_size"] += 1
        if source_x0 is None:
            source_x0 = np.zeros_like(phi0)

    if score_list["tilt-y"]:
        card["iqmc"]["krylov_vector_size"] += 1
        if source_y0 is None:
            source_y0 = np.zeros_like(phi0)

    if score_list["tilt-z"]:
        card["iqmc"]["krylov_vector_size"] += 1
        if source_z0 is None:
            source_z0 = np.zeros_like(phi0)

    if score_list["tilt-xy"]:
        card["iqmc"]["krylov_vector_size"] += 1
        if source_xy0 is None:
            source_xy0 = np.zeros_like(phi0)

    if score_list["tilt-xz"]:
        card["iqmc"]["krylov_vector_size"] += 1
        if source_xz0 is None:
            source_xz0 = np.zeros_like(phi0)

    if score_list["tilt-yz"]:
        card["iqmc"]["krylov_vector_size"] += 1
        if source_yz0 is None:
            source_yz0 = np.zeros_like(phi0)

    if fission_source0 is not None:
        card["iqmc"]["score"]["fission-source"] = fission_source0

    card["iqmc"]["score"]["flux"] = phi0
    card["iqmc"]["score"]["tilt-x"] = source_x0
    card["iqmc"]["score"]["tilt-y"] = source_y0
    card["iqmc"]["score"]["tilt-z"] = source_z0
    card["iqmc"]["score"]["tilt-xy"] = source_xy0
    card["iqmc"]["score"]["tilt-xz"] = source_xz0
    card["iqmc"]["score"]["tilt-yz"] = source_yz0
    card["iqmc"]["source"] = source0
    card["iqmc"]["fixed_source"] = fixed_source
    card["iqmc"]["fixed_source_solver"] = fixed_source_solver
    card["iqmc"]["eigenmode_solver"] = eigenmode_solver
    card["iqmc"]["preconditioner_sweeps"] = preconditioner_sweeps
    card["iqmc"]["krylov_restart"] = krylov_restart


def weight_roulette(chance, wr_threshold):
    """
    If neutron weight is below wr_threshold, then enter weight rouelette
    technique. Neutron has 'chance' probability of having its weight increased
    by factor of 1/CHANCE, and 1-CHANCE probability of terminating.

    Parameters
    ----------
    chance : probability of survival
    wr_threshold : weight_roulette() is called on a particle
                    if P['w'] <= wr_threshold

    Returns
    -------
    None.

    """
    card = mcdc.input_deck.technique
    card["weight_roulette"] = True
    card["wr_chance"] = chance
    card["wr_threshold"] = wr_threshold


def IC_generator(
    N_neutron=0,
    N_precursor=0,
    cycle_stretch=1.0,
    neutron_density=None,
    max_neutron_density=None,
    precursor_density=None,
    max_precursor_density=None,
):
    """
    Turn on initial condition generator, which samples initial neutrons and precursors
    during an eigenvalue simulation.


    Parameters
    ----------
    N_neutron : int
        Neutron target size
    N_precursor : int
        Delayed neutron precursot target size
    cycle_stretch : float
        Factor to strethch number of cycles. Higher cycle stretch reduces inter-cycle
        correlation.
    neutron_density, max_neutron_density : float
        Total and maximum neutron density, required if `N_neutron` > 0.
    precursor_density, max_precursor_density : float
        Total and maximum precursor density, required if `N_precursor` > 0.
    """

    # Turn on eigenmode and population control
    eigenmode()
    population_control()

    # Set parameters
    card = mcdc.input_deck.technique
    card["IC_generator"] = True
    card["IC_N_neutron"] = N_neutron
    card["IC_N_precursor"] = N_precursor

    # Setting parameters
    card_setting = mcdc.input_deck.setting
    N_particle = card_setting["N_particle"]

    # Check optional parameters
    if N_neutron > 0.0:
        if neutron_density is None or max_neutron_density is None:
            print_error("IC generator requires neutron_density and max_neutron_density")
        card["IC_neutron_density"] = N_particle * neutron_density
        card["IC_neutron_density_max"] = max_neutron_density
    if N_precursor > 0.0:
        if precursor_density is None:
            print_error(
                "IC generator requires precursor_density and max_precursor_density"
            )
        card["IC_precursor_density"] = N_particle * precursor_density
        card["IC_precursor_density_max"] = max_precursor_density

    # Set number of active cycles
    n = card["IC_neutron_density"]
    n_max = card["IC_neutron_density_max"]
    C = card["IC_precursor_density"]
    C_max = card["IC_precursor_density_max"]
    N_cycle1 = 0.0
    N_cycle2 = 0.0
    if N_neutron > 0:
        N_cycle1 = math.ceil(cycle_stretch * math.ceil(n_max / n * N_neutron))
    if N_precursor > 0:
        N_cycle2 = math.ceil(cycle_stretch * math.ceil(C_max / C * N_precursor))
    N_cycle = max(N_cycle1, N_cycle2)
    card_setting["N_cycle"] = N_cycle
    card_setting["N_active"] = N_cycle


def dsm(order=1):
    card = mcdc.input_deck.technique
    if order > 2:
        print_error("DSM currently only supports up to second-order sensitivities")
    card["dsm_order"] = order


def uq(**kw):
    def append_card(delta_card, global_tag):
        delta_card["distribution"] = dist
        delta_card["flags"] = []
        for key in kw.keys():
            check_support(parameter["tag"] + " parameter", key, parameter_list, False)
            delta_card["flags"].append(key)
            delta_card[key] = kw[key]
        mcdc.input_deck.uq_deltas[global_tag].append(delta_card)

    mcdc.input_deck.technique["uq"] = True
    # Make sure N_batch > 1
    if mcdc.input_deck.setting["N_batch"] <= 1:
        print_error("Must set N_batch>1 with mcdc.setting() prior to mcdc.uq() call.")

    # Check uq parameter
    parameter_ = check_support(
        "uq parameter",
        list(kw)[0],
        ["nuclide", "material", "surface", "source"],
        False,
    )
    parameter = kw[parameter_]
    del kw[parameter_]
    parameter["uq"] = True

    # Confirm supplied distribution
    check_requirement("uq", kw, ["distribution"])
    dist = check_support("distribution", kw["distribution"], ["uniform"], False)
    del kw["distribution"]

    # Only remaining keywords should be the parameter delta(s)

    if parameter["tag"] == "Material":
        parameter_list = [
            "capture",
            "scatter",
            "fission",
            "nu_s",
            "nu_p",
            "nu_d",
            "chi_p",
            "chi_d",
            "speed",
            "decay",
        ]
        global_tag = "materials"
        if parameter["N_nuclide"] == 1:
            nuc_card = make_card_nuclide(parameter["G"], parameter["J"])
            nuc_card["ID"] = parameter["nuclide_IDs"][0]
            append_card(nuc_card, "nuclides")
        delta_card = make_card_material(
            parameter["N_nuclide"], parameter["G"], parameter["J"]
        )
        for name in ["ID", "nuclide_IDs", "nuclide_densities"]:
            delta_card[name] = parameter[name]
    elif parameter["tag"] == "Nuclide":
        parameter_list = [
            "capture",
            "scatter",
            "fission",
            "nu_s",
            "nu_p",
            "nu_d",
            "chi_p",
            "chi_d",
            "speed",
            "decay",
        ]
        global_tag = "nuclides"
        delta_card = make_card_nuclide(parameter["G"], parameter["J"])
        delta_card["ID"] = parameter["ID"]
    append_card(delta_card, global_tag)
    # elif parameter['tag'] is 'Surface':
    # elif parameter['tag'] is 'Source':


# ==============================================================================
# Util
# ==============================================================================


def nuclide_registered(name):
    for card in mcdc.input_deck.nuclides:
        if name == card["name"]:
            return True
    return False


def get_nuclide(name):
    for card in mcdc.input_deck.nuclides:
        if name == card["name"]:
            return card


def print_card(card):
    if isinstance(card, SurfaceHandle):
        card = card.card
    for key in card:
        if key == "tag":
            print(card[key] + " card")
        else:
            print("  " + key + " : " + str(card[key]))


def check_support(label, value, supported, replace=True):
    if replace:
        value = value.replace("_", "-").replace(" ", "-").lower()
    supported_str = "{"
    for str_ in supported:
        supported_str += str_ + ", "
    supported_str = supported_str[:-2] + "}"
    if value not in supported:
        print_error("Unsupported " + label + ": " + value + "\n" + supported_str)
    return value


def check_requirement(label, kw, required):
    missing = "{"
    error = False
    for req in required:
        if req not in kw.keys():
            error = True
            missing += req + ", "
    missing = missing[:-2] + "}"
    if error:
        print_error("Parameters " + missing + " are required for" + label)


# ==============================================================================
# Reset
# ==============================================================================


def reset_cards():
    mcdc.input_deck.reset()



import matplotlib.pyplot as plt
import h5py, sys
import numpy as np


output = sys.argv[1]

# Load results
with h5py.File(output, "r") as f:
    # Grid
    x = f["tally/grid/x"][:]
    dx = x[1:] - x[:-1]
    x_mid = 0.5 * (x[:-1] + x[1:])
    # Fluks
    phi_fast = f["tally/flux/mean"][1, :] / dx
    phi_fast_sd = f["tally/flux/sdev"][1, :] / dx
    phi_thermal = f["tally/flux/mean"][0, :] / dx
    phi_thermal_sd = f["tally/flux/sdev"][0, :] / dx

# Plot results
fig, ax1 = plt.subplots(figsize=(5, 3))
ax2 = ax1.twinx()

y = phi_fast
y_sd = phi_fast_sd
p1 = ax1.step(x_mid, y, "-b", where="mid", label="Fast flux")
ax1.fill_between(x_mid, y - y_sd, y + y_sd, alpha=0.2, color="b", step="mid")

y = phi_thermal
y_sd = phi_thermal_sd
p2 = ax2.step(x_mid, y, "--r", where="mid", label="Thermal flux")
ax2.fill_between(x_mid, y - y_sd, y + y_sd, alpha=0.2, color="r", step="mid")

ax1.set_xlabel(r"$x$ [cm]")
ax1.set_ylabel("Fast flux [/cm-s]", color="b")
ax1.tick_params(axis="y", colors="b")
ax2.set_ylabel("Thermal flux [/cm-s]", color="r")
ax2.tick_params(axis="y", colors="r")

"""
lim1 = ax1.get_ylim()
lim2 = ax2.get_ylim()
max_range = max(lim1[1]-lim1[0], lim2[1]-lim2[0])
ax1.set_ylim(top=lim1[0]+max_range)
ax2.set_ylim(top=lim2[0]+max_range)
"""

ax1.legend(handles=p1 + p2, loc="lower left")
plt.show()



import numpy as np
from scipy.integrate import quad


# =============================================================================
# Reference solution generator
# =============================================================================

# Scattering ratio
c = 1.1
i = complex(0, 1)

# Spatial grid
J = 201
x = np.linspace(-20.5, 20.5, J + 1)

# Time grid
K = 20
t = np.linspace(0.0, 20.0, K + 1)


def integrand(u, eta, t):
    q = (1 + eta) / (1 - eta)
    xi = (np.log(q) + i * u) / (eta + i * np.tan(u / 2))
    return (
        1.0
        / (np.cos(u / 2)) ** 2
        * (xi**2 * np.e ** (c * t / 2 * (1 - eta**2) * xi)).real
    )


def phi(x, t):
    if t == 0.0 or abs(x) >= t:
        return 0.0
    eta = x / t
    integral = quad(integrand, 0.0, np.pi, args=(eta, t))[0]
    return np.e**-t / 2 / t * (1 + c * t / 4 / np.pi * (1 - eta**2) * integral)


def phi_t(t, x):
    if t == 0.0 or abs(x) >= t:
        return 0.0
    eta = x / t
    integral = quad(integrand, 0.0, np.pi, args=(eta, t))[0]
    return np.e**-t / 2 / t * (1 + c * t / 4 / np.pi * (1 - eta**2) * integral)


def phiX(x, t0, t1):
    return quad(phi_t, t0, t1, args=(x))[0]


phi_avg = np.zeros([K, J])

for k in range(K):
    for j in range(J):
        x0 = x[j]
        x1 = x[j + 1]
        dx = x1 - x0
        t0 = t[k]
        t1 = t[k + 1]
        dt = t1 - t0
        phi_avg[k, j] = quad(phiX, x0, x1, args=(t0, t1))[0] / dx / dt

for j in range(J + 1):
    for k in range(K):
        t0 = t[k]
        t1 = t[k + 1]
        dt = t1 - t0

phi_avg = np.nan_to_num(phi_avg)

np.savez("reference.npz", x=x, t=t, phi=phi_avg)



import h5py
import numpy as np

import mcdc

# =============================================================================
# Materials
# =============================================================================

# Load material data
lib = h5py.File("../../c5g7_xs.h5", "r")


# Setter
def set_mat(mat):
    return mcdc.material(
        capture=mat["capture"][:],
        scatter=mat["scatter"][:],
        fission=mat["fission"][:],
        nu_p=mat["nu_p"][:],
        nu_d=mat["nu_d"][:],
        chi_p=mat["chi_p"][:],
        chi_d=mat["chi_d"][:],
        speed=mat["speed"],
        decay=mat["decay"],
    )


# Set the material
mat_uo2 = set_mat(lib["uo2"])  # Fuel: UO2
mat_mox43 = set_mat(lib["mox43"])  # Fuel: MOX 4.3%
mat_mox7 = set_mat(lib["mox7"])  # Fuel: MOX 7.0%
mat_mox87 = set_mat(lib["mox87"])  # Fuel: MOX 8.7%
mat_gt = set_mat(lib["gt"])  # Guide tube
mat_fc = set_mat(lib["fc"])  # Fission chamber
mat_cr = set_mat(lib["cr"])  # Control rod
mat_mod = set_mat(lib["mod"])  # Moderator

# =============================================================================
# Pin cells
# =============================================================================

pitch = 1.26
radius = 0.54
core_height = 128.52
refl_thick = 21.42

# Control rod banks fractions
#   All out: 0.0
#   All in : 1.0
cr1 = 0.0
cr2 = 0.0
cr3 = 0.0
cr4 = 0.0
# Control rod banks interfaces
cr1 = core_height * (0.5 - cr1)
cr2 = core_height * (0.5 - cr2)
cr3 = core_height * (0.5 - cr3)
cr4 = core_height * (0.5 - cr4)

# Surfaces
cy = mcdc.surface("cylinder-z", center=[0.0, 0.0], radius=radius)
z1 = mcdc.surface("plane-z", z=cr1)  # Control rod banks interfaces
z2 = mcdc.surface("plane-z", z=cr2)
z3 = mcdc.surface("plane-z", z=cr3)
z4 = mcdc.surface("plane-z", z=cr4)
zf = mcdc.surface("plane-z", z=core_height / 2)

# Fission chamber
fc = mcdc.cell([-cy], mat_fc)
mod = mcdc.cell([+cy], mat_mod)
fission_chamber = mcdc.universe([fc, mod])["ID"]

# Fuel rods
uo2 = mcdc.cell([-cy, -zf], mat_uo2)
mox4 = mcdc.cell([-cy, -zf], mat_mox43)
mox7 = mcdc.cell([-cy, -zf], mat_mox7)
mox8 = mcdc.cell([-cy, -zf], mat_mox87)
moda = mcdc.cell([-cy, +zf], mat_mod)  # Water above pin
fuel_uo2 = mcdc.universe([uo2, mod, moda])["ID"]
fuel_mox43 = mcdc.universe([mox4, mod, moda])["ID"]
fuel_mox7 = mcdc.universe([mox7, mod, moda])["ID"]
fuel_mox87 = mcdc.universe([mox8, mod, moda])["ID"]

# Control rods and guide tubes
cr1 = mcdc.cell([-cy, +z1], mat_cr)
cr2 = mcdc.cell([-cy, +z2], mat_cr)
cr3 = mcdc.cell([-cy, +z3], mat_cr)
cr4 = mcdc.cell([-cy, +z4], mat_cr)
gt1 = mcdc.cell([-cy, -z1], mat_gt)
gt2 = mcdc.cell([-cy, -z2], mat_gt)
gt3 = mcdc.cell([-cy, -z3], mat_gt)
gt4 = mcdc.cell([-cy, -z4], mat_gt)
control_rod1 = mcdc.universe([cr1, gt1, mod])["ID"]
control_rod2 = mcdc.universe([cr2, gt2, mod])["ID"]
control_rod3 = mcdc.universe([cr3, gt3, mod])["ID"]
control_rod4 = mcdc.universe([cr4, gt4, mod])["ID"]

# =============================================================================
# Fuel lattices
# =============================================================================

# UO2 lattice 1
u = fuel_uo2
c = control_rod1
f = fission_chamber
lattice_1 = mcdc.lattice(
    x=[-pitch * 17 / 2, pitch, 17],
    y=[-pitch * 17 / 2, pitch, 17],
    universes=[
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, u, u, c, u, u, c, u, u, c, u, u, u, u, u],
        [u, u, u, c, u, u, u, u, u, u, u, u, u, c, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, c, u, u, c, u, u, c, u, u, c, u, u, c, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, c, u, u, c, u, u, f, u, u, c, u, u, c, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, c, u, u, c, u, u, c, u, u, c, u, u, c, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, c, u, u, u, u, u, u, u, u, u, c, u, u, u],
        [u, u, u, u, u, c, u, u, c, u, u, c, u, u, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
    ],
)

# MOX lattice 2
l = fuel_mox43
m = fuel_mox7
n = fuel_mox87
c = control_rod2
f = fission_chamber
lattice_2 = mcdc.lattice(
    x=[-pitch * 17 / 2, pitch, 17],
    y=[-pitch * 17 / 2, pitch, 17],
    universes=[
        [l, l, l, l, l, l, l, l, l, l, l, l, l, l, l, l, l],
        [l, m, m, m, m, m, m, m, m, m, m, m, m, m, m, m, l],
        [l, m, m, m, m, c, m, m, c, m, m, c, m, m, m, m, l],
        [l, m, m, c, m, n, n, n, n, n, n, n, m, c, m, m, l],
        [l, m, m, m, n, n, n, n, n, n, n, n, n, m, m, m, l],
        [l, m, c, n, n, c, n, n, c, n, n, c, n, n, c, m, l],
        [l, m, m, n, n, n, n, n, n, n, n, n, n, n, m, m, l],
        [l, m, m, n, n, n, n, n, n, n, n, n, n, n, m, m, l],
        [l, m, c, n, n, c, n, n, f, n, n, c, n, n, c, m, l],
        [l, m, m, n, n, n, n, n, n, n, n, n, n, n, m, m, l],
        [l, m, m, n, n, n, n, n, n, n, n, n, n, n, m, m, l],
        [l, m, c, n, n, c, n, n, c, n, n, c, n, n, c, m, l],
        [l, m, m, m, n, n, n, n, n, n, n, n, n, m, m, m, l],
        [l, m, m, c, m, n, n, n, n, n, n, n, m, c, m, m, l],
        [l, m, m, m, m, c, m, m, c, m, m, c, m, m, m, m, l],
        [l, m, m, m, m, m, m, m, m, m, m, m, m, m, m, m, l],
        [l, l, l, l, l, l, l, l, l, l, l, l, l, l, l, l, l],
    ],
)

# MOX lattice 3
l = fuel_mox43
m = fuel_mox7
n = fuel_mox87
c = control_rod3
f = fission_chamber
lattice_3 = mcdc.lattice(
    x=[-pitch * 17 / 2, pitch, 17],
    y=[-pitch * 17 / 2, pitch, 17],
    universes=[
        [l, l, l, l, l, l, l, l, l, l, l, l, l, l, l, l, l],
        [l, m, m, m, m, m, m, m, m, m, m, m, m, m, m, m, l],
        [l, m, m, m, m, c, m, m, c, m, m, c, m, m, m, m, l],
        [l, m, m, c, m, n, n, n, n, n, n, n, m, c, m, m, l],
        [l, m, m, m, n, n, n, n, n, n, n, n, n, m, m, m, l],
        [l, m, c, n, n, c, n, n, c, n, n, c, n, n, c, m, l],
        [l, m, m, n, n, n, n, n, n, n, n, n, n, n, m, m, l],
        [l, m, m, n, n, n, n, n, n, n, n, n, n, n, m, m, l],
        [l, m, c, n, n, c, n, n, f, n, n, c, n, n, c, m, l],
        [l, m, m, n, n, n, n, n, n, n, n, n, n, n, m, m, l],
        [l, m, m, n, n, n, n, n, n, n, n, n, n, n, m, m, l],
        [l, m, c, n, n, c, n, n, c, n, n, c, n, n, c, m, l],
        [l, m, m, m, n, n, n, n, n, n, n, n, n, m, m, m, l],
        [l, m, m, c, m, n, n, n, n, n, n, n, m, c, m, m, l],
        [l, m, m, m, m, c, m, m, c, m, m, c, m, m, m, m, l],
        [l, m, m, m, m, m, m, m, m, m, m, m, m, m, m, m, l],
        [l, l, l, l, l, l, l, l, l, l, l, l, l, l, l, l, l],
    ],
)

# UO2 lattice 4
u = fuel_uo2
c = control_rod4
f = fission_chamber
lattice_4 = mcdc.lattice(
    x=[-pitch * 17 / 2, pitch, 17],
    y=[-pitch * 17 / 2, pitch, 17],
    universes=[
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, u, u, c, u, u, c, u, u, c, u, u, u, u, u],
        [u, u, u, c, u, u, u, u, u, u, u, u, u, c, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, c, u, u, c, u, u, c, u, u, c, u, u, c, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, c, u, u, c, u, u, f, u, u, c, u, u, c, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, c, u, u, c, u, u, c, u, u, c, u, u, c, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, c, u, u, u, u, u, u, u, u, u, c, u, u, u],
        [u, u, u, u, u, c, u, u, c, u, u, c, u, u, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
        [u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u, u],
    ],
)

# =============================================================================
# Assemblies and core
# =============================================================================

# Surfaces
x0 = mcdc.surface("plane-x", x=0.0, bc="reflective")
x1 = mcdc.surface("plane-x", x=pitch * 17)
x2 = mcdc.surface("plane-x", x=pitch * 17 * 2)
x3 = mcdc.surface("plane-x", x=pitch * 17 * 3, bc="vacuum")

y0 = mcdc.surface("plane-y", y=-pitch * 17 * 3, bc="vacuum")
y1 = mcdc.surface("plane-y", y=-pitch * 17 * 2)
y2 = mcdc.surface("plane-y", y=-pitch * 17)
y3 = mcdc.surface("plane-y", y=0.0, bc="reflective")

z0 = mcdc.surface("plane-z", z=-(core_height / 2 + refl_thick), bc="vacuum")
z1 = mcdc.surface("plane-z", z=-(core_height / 2))
z2 = mcdc.surface("plane-z", z=(core_height / 2 + refl_thick), bc="vacuum")

# Assembly cells
center = np.array([pitch * 17 / 2, -pitch * 17 / 2, 0.0])
assembly_1 = mcdc.cell([+x0, -x1, +y2, -y3, +z1, -z2], lattice_1, lattice_center=center)

center += np.array([pitch * 17, 0.0, 0.0])
assembly_2 = mcdc.cell([+x1, -x2, +y2, -y3, +z1, -z2], lattice_2, lattice_center=center)

center += np.array([-pitch * 17, -pitch * 17, 0.0])
assembly_3 = mcdc.cell([+x0, -x1, +y1, -y2, +z1, -z2], lattice_3, lattice_center=center)

center += np.array([pitch * 17, 0.0, 0.0])
assembly_4 = mcdc.cell([+x1, -x2, +y1, -y2, +z1, -z2], lattice_4, lattice_center=center)

# Bottom reflector cell
reflector_bottom = mcdc.cell([+x0, -x3, +y0, -y3, +z0, -z1], mat_mod)

# Side reflectors
reflector_south = mcdc.cell([+x0, -x3, +y0, -y1, +z1, -z2], mat_mod)
reflector_east = mcdc.cell([+x2, -x3, +y1, -y3, +z1, -z2], mat_mod)

# Root universe
mcdc.universe(
    [
        assembly_1,
        assembly_2,
        assembly_3,
        assembly_4,
        reflector_bottom,
        reflector_south,
        reflector_east,
    ],
    root=True,
)

# =============================================================================
# Set source
# =============================================================================
# Uniform in energy

source = mcdc.source(
    x=[0.0, pitch * 17 * 2],
    y=[-pitch * 17 * 2, 0.0],
    z=[-core_height / 2, core_height / 2],
    energy=np.ones(7),
)

# =============================================================================
# Set tally and parameter, and then run mcdc
# =============================================================================

# Tally
x_grid = np.linspace(0.0, pitch * 17 * 3, 17 * 3 + 1)
y_grid = np.linspace(-pitch * 17 * 3, 0.0, 17 * 3 + 1)
z_grid = np.linspace(
    -(core_height / 2 + refl_thick), (core_height / 2 + refl_thick), 102 + 17 * 2 + 1
)
g_grid = np.array([-0.5, 3.5, 6.5])  # Collapsing to fast (1-4) and slow (5-7)

mcdc.tally(scores=["flux"], x=x_grid, y=y_grid, z=z_grid, g=g_grid)

# Setting
mcdc.setting(N_particle=1e3)
mcdc.eigenmode(N_inactive=50, N_active=150, gyration_radius="all")
mcdc.population_control()

# Run
mcdc.run()



import matplotlib.pyplot as plt
import h5py
import numpy as np


# Load result
with h5py.File("output.h5", "r") as f:
    x = f["tally/grid/x"][:]
    dx = [x[1:] - x[:-1]][-1]
    x_mid = 0.5 * (x[:-1] + x[1:])

    phi = f["tally/flux/mean"][:]
    phi_sd = f["tally/flux/sdev"][:]


# Plot result
X, Y = np.meshgrid(x_mid, x_mid)
Z = phi
Z = np.log10(np.abs(Z))
fig, ax = plt.subplots(subplot_kw={"projection": "3d"})
ax.plot_surface(Y, X, Z, edgecolor="k", color="white")
ax.set_xlabel("x")
ax.set_ylabel("y")
ax.zaxis.set_rotate_label(False)
ax.set_zlabel("Log10 of scalar flux", rotation=90)
ax.view_init(elev=18, azim=38)
plt.show()

Z = phi_sd / phi
Z = np.log10(np.abs(Z))
fig, ax = plt.subplots(subplot_kw={"projection": "3d"})
ax.plot_surface(Y, X, Z, edgecolor="k", color="white")
ax.set_xlabel("x")
ax.set_ylabel("y")
ax.zaxis.set_rotate_label(False)
ax.set_zlabel("Log10 of scalar flux rel. stdev.", rotation=90)
ax.view_init(elev=18, azim=38)
plt.show()



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
import h5py
import matplotlib.animation as animation


# =============================================================================
# Plot results
# =============================================================================

# Results
with h5py.File("output.h5", "r") as f:
    x = f["tally/grid/x"][:]
    x_mid = 0.5 * (x[:-1] + x[1:])
    y = f["tally/grid/y"][:]
    y_mid = 0.5 * (y[:-1] + y[1:])
    t = f["tally/grid/t"][:]
    t_mid = 0.5 * (t[:-1] + t[1:])
    X, Y = np.meshgrid(y, x)

    phi = f["tally/flux/mean"][:]
    phi_sd = f["tally/flux/sdev"][:]

fig, ax = plt.subplots()
cax = ax.pcolormesh(X, Y, phi[0], vmin=phi[0].min(), vmax=phi[0].max())
text = ax.text(0.02, 1.02, "", transform=ax.transAxes)
ax.set_aspect("equal", "box")
ax.set_xlabel("$y$ [cm]")
ax.set_ylabel("$x$ [cm]")
ax.set_aspect("equal")


def animate(i):
    cax.set_array(phi[i])
    cax.set_clim(phi[i].min(), phi[i].max())
    text.set_text(r"$t \in [%.1f,%.1f]$ s" % (t[i], t[i + 1]))


anim = animation.FuncAnimation(fig, animate, interval=10, frames=len(t) - 1)
plt.show()
